(*  Title:      CyclesOfFutures.thy
    ID:         $Id: CyclesOfFutures.thy,v 1.5 2008/11/14 14:09:04 sowilo Exp $
    Author:     Ludovic Henrio and Florian Kammuller
                2008

    Note:       proof of absence of cycle of futures
*)

header {* Absence of cycle of futures for ASP_fun *}

theory CyclesOfFutures imports WellFormedness begin


datatype Reference =
    RefAct ActivityRef
  | RefFut FutureRef 

(* defines the knowledge relationship between AO and future:
  a fut f knows a AO A (resp Fut f') if A (resp f') it is referenced 
          by the request corresponding to the fut f
  a AO A knows  a AO A (resp Fut f') if A (resp f') it is referenced 
          by the active object of A *)
constdefs 
  fcontainsf :: "configuration \<Rightarrow> (FutureRef * FutureRef) set"
  "(fcontainsf c)=={(f, f'). \<exists> A\<in>(dom c). \<exists> (E::general_context). 
                            (f\<in> dom (c.\<^sub>RA) \<and>(c.\<^sub>FA<f>) = (E\<up>(FutRef(f'))))}"

  fcontainsa :: "configuration \<Rightarrow> (FutureRef * ActivityRef) set"
  "(fcontainsa c)== { (f, B). \<exists> A\<in>(dom c). \<exists> (E::general_context). 
                             (f\<in> dom (c.\<^sub>RA) \<and>(c.\<^sub>FA<f>)=(E\<up>(ActRef(B))))}"

  acontainsf :: "configuration \<Rightarrow> (ActivityRef * FutureRef) set"
  "(acontainsf c)=={(A, f'). A\<in> dom c \<and> (\<exists> (E::general_context). (c.\<^sub>AA) = (E\<up>(FutRef(f'))))}"

  acontainsa :: "configuration \<Rightarrow> (ActivityRef * ActivityRef) set"
  "(acontainsa c)== {(A, B). A\<in> dom c \<and> (\<exists> (E::general_context). (c.\<^sub>AA) = E\<up>(ActRef(B)))}"
constdefs
  knows :: "configuration \<Rightarrow> (Reference * Reference) set"
  " (knows c) == {(RefFut f,RefFut f')| f f'. (\<exists> A\<in>(dom c). \<exists> (E::general_context). 
                            (f\<in> dom (c.\<^sub>RA) \<and>(c.\<^sub>FA<f>) = (E\<up>(FutRef(f')))))} \<union> 
              {(RefFut f, RefAct B)| f B. \<exists> A\<in>(dom c). \<exists> (E::general_context). 
                             (f\<in> dom (c.\<^sub>RA) \<and>(c.\<^sub>FA<f>)=(E\<up>(ActRef(B))))}\<union>
              {(RefAct A, RefFut f')| A f'. A\<in> dom c \<and> (\<exists> (E::general_context). 
                              (c.\<^sub>AA) = (E\<up>(FutRef(f'))))} \<union>
              {(RefAct A, RefAct B)| A B. A\<in> dom c \<and> (\<exists> (E::general_context). (c.\<^sub>AA) = E\<up>(ActRef(B)))}"

(* now the transitive clusure, but remembering the followed path *)
inductive_set 
  knowsrec:: "configuration \<Rightarrow> (Reference * Reference * Reference list) set"
  for c :: "configuration"
where 
   krDirect: "(r,r') \<in> (knows c) \<Longrightarrow>(r,r',Nil)\<in>(knowsrec c)"
 | krRec: "(r,r')\<in>(knows c) \<and>(r',r'',rL) \<in> (knowsrec c) \<Longrightarrow>(r,r'',r'#rL)\<in>(knowsrec c)"


lemma pre_knowsrecI[rule_format]: "  \<forall> r r'. ((r,r',rL)\<in>(knowsrec c) \<longrightarrow>(r,r')\<in>((knows c)\<^sup>+))"
apply (induct_tac rL);
apply ((rule allI)+,rule impI);
apply (force elim: knowsrec.cases);

apply ((rule allI)+,rule impI);
apply (erule knowsrec.cases);
apply simp;
apply force;
done;

lemma knowsrec_rev[rule_format]:
  " \<forall> r r' r''. (r,r',rL) \<in> (knowsrec c)\<and>(r',r'')\<in>(knows c) \<longrightarrow>(r,r'',rL@[r'])\<in>(knowsrec c)"
apply (induct_tac rL)
apply (auto intro: knowsrec.intros elim: knowsrec.cases) 
apply (erule knowsrec.cases);
apply simp+;
apply (force intro: knowsrec.intros elim: knowsrec.cases) 
done;

(* FORMATTED:
   (?r, ?r', ?rL) \<in> knowsrec ?c \<and> (?r', ?r'') \<in> knows ?c \<Longrightarrow> (?r, ?r'', ?rL @ [?r']) \<in> knowsrec ?c*)

lemma knowsreccases_rev[rule_format]:
  " \<forall> r r' r''. (r,r'',rL@[r'])\<in>(knowsrec c) \<longrightarrow>(r,r',rL) \<in> (knowsrec c)\<and>(r',r'')\<in>(knows c)"
apply (induct_tac rL)
apply (auto intro: knowsrec.intros elim: knowsrec.cases) 
apply (erule knowsrec.cases);
apply simp+;
apply (force intro: knowsrec.intros elim: knowsrec.cases) 
done;

(* FORMATTED:
   (?r, ?r'', ?rL @ [?r']) \<in> knowsrec ?c \<Longrightarrow> (?r, ?r', ?rL) \<in> knowsrec ?c \<and> (?r', ?r'') \<in> knows ?c
*)

lemma knowsrecI[rule_format]: " (r,r')\<in>((knows c)\<^sup>+) = (\<exists> rL. (r,r',rL)\<in>(knowsrec c))"
apply (rule iffI);
apply (erule trancl_induct);
apply (force intro: knowsrec.intros )
apply (erule exE);
apply (rule_tac x= "rL@[y]" in exI)
apply (force intro: knowsrec.intros knowsrec_rev)
apply (erule exE);
apply (rule_tac P="(r, r', rL) \<in> knowsrec c" in mp);
apply (force simp: pre_knowsrecI);
apply assumption;
done;


(* now we can define absene of cycle easily *)
constdefs
  nocycle :: "configuration \<Rightarrow>bool"
   "nocycle c == \<forall> r. ((r,r)\<notin> ((knows c)\<^sup>+))"

lemma noRefAct2: " noRef a \<Longrightarrow> \<forall> E A. (a\<noteq>(E::general_context)\<up>ActRef A)"
apply (auto simp: noRef_act);
done

lemma noRefFut2: " noRef a \<Longrightarrow> \<forall> E f. (a\<noteq>(E::general_context)\<up>FutRef f)"
apply (auto simp: noRef_fut);
done

lemma decom_knowsrec: "(r,r')\<in>(knows c)\<^sup>+ = (\<exists> r''. ((r,r'')\<in>(knows c) \<and>(r''=r' \<or> (r'',r')\<in>(knows c)\<^sup>+)))"
apply auto;
apply (erule trancl_induct)
apply auto;
apply (rule_tac x=r'' in exI);
apply (auto intro: trancl.intros);
done;

(* no cycle in the init_config *)
lemma initial_nocycle: "(noRef a) \<Longrightarrow>nocycle (init_config a)"
apply (unfold nocycle_def init_config_def);
apply (rule allI);
apply (simp add: knowsrecI); 
apply (rule allI);
apply (induct_tac rL);
apply auto;
apply (erule knowsrec.cases);
apply (unfold knows_def);
apply (auto simp: ActiveObject_def OneRequest_def RequestQueue_def noRef_fut noRef_act);
apply (case_tac "A=aref 0");
apply (case_tac "f'=fref 0");apply (force simp: noRef_fut);
apply (force simp: noRef_fut);apply (force simp: noRef_fut);

apply (case_tac "B=aref 0");
apply simp;
apply (case_tac E);
apply (auto simp: noRef_fut noRef_act);
apply (rotate_tac 1,drule sym, simp add: map_upd_nonempty)

apply (rotate_tac -1,erule knowsrec.cases);
apply force;
apply (simp+,(erule conjE)+);
apply (unfold knows_def);
apply (auto simp: ActiveObject_def OneRequest_def RequestQueue_def noRef_fut noRef_act);
apply (case_tac "A=aref 0");
apply (case_tac "f=fref 0");apply (force simp: noRef_fut);
apply (force simp: noRef_fut);apply (force simp: noRef_fut);
apply (case_tac "A=aref 0");
apply (case_tac "f=fref 0");apply (force simp: noRef_act);
apply (force simp: noRef_act);apply (force simp: noRef_act);

apply (case_tac "A=aref 0");
apply simp;
apply (case_tac E);
apply (auto simp: noRef_fut noRef_act);
apply (rotate_tac 1,drule sym, simp add: map_upd_nonempty)
apply (case_tac "A=aref 0");
apply simp;
apply (case_tac E);
apply (auto simp: noRef_fut noRef_act);
apply (rotate_tac 1,drule sym, simp add: map_upd_nonempty)
done;

lemma subtrancl: "A\<subseteq>A' \<Longrightarrow> A\<^sup>+\<subseteq>A'\<^sup>+"
apply auto;
apply (erule trancl_induct);
apply auto;
done;

lemma subcontains_nocycle: "\<lbrakk>nocycle c; knows c'\<subseteq> knows c\<rbrakk>\<Longrightarrow>nocycle c'"
apply (unfold nocycle_def);
apply (drule subtrancl);
apply (auto);
done;

(* first references in the already existing active objects are unchanged by red *)
lemma red_RefAct: "c \<rightarrow>\<^sub>\<zeta> c'\<Longrightarrow> \<forall>A \<in> dom c. ((RefAct A,r) \<in> knows c) = ((RefAct A,r)\<in> knows c')"
 apply (subgoal_tac " \<forall> A \<in> dom c.(c.\<^sub>AA)=(c'.\<^sub>AA)");
 apply (frule actdom_and_red_inc);
 apply (unfold knows_def  RequestQueue_def ActiveObject_def OneRequest_def);
 apply auto;
 apply force;
 apply (rule_tac x=E in exI);
 apply force;
 apply force;
 apply (drule_tac x=E in spec);
 apply simp+;
 apply (frule_tac P=" snd (the (c A)) = snd (the (c' A))" in zeta.cases);
 apply auto;
done;

(* consequently, if there is no created AO ... *)
lemma red_RefAct2: "\<lbrakk>c  \<rightarrow>\<^sub>\<zeta> c';dom c = dom c'\<rbrakk>\<Longrightarrow> ((RefAct A,r) \<in> knows c) = ((RefAct A,r)\<in> knows c')"
apply (drule_tac r=r in red_RefAct);
apply auto;
apply (drule_tac x=A in bspec);
apply (unfold knows_def)
apply (force);
apply force;
apply (drule_tac x=A in bspec);
apply auto;
done;

lemma list_set_decomp[rule_format]: "(r \<in>set rL) \<longrightarrow>(\<exists> rL1 rL2 . rL = rL1@r#rL2)"
apply (induct_tac rL);
apply auto;
apply (rule_tac x="a # rL1" in exI);
apply force;
done;

(* how to decompose a knowsrec path in the "middle" *)
lemma decomp_knowsrec[rule_format]: 
    " \<forall> r1 r2. (r1,r2,rL@r'#rL')\<in>knowsrec c \<longrightarrow> ((r1,r',rL)\<in>knowsrec c \<and> (r',r2,rL')\<in>knowsrec c)"
apply (induct_tac rL')
apply (induct_tac rL)
apply simp+;
apply ((rule allI)+,rule impI,  rule conjI);
apply (force intro: knowsrec.intros elim: knowsrec.cases) 
apply (force intro: knowsrec.intros elim: knowsrec.cases) 
apply ((rule allI)+,rule impI,  rule conjI);
apply (force intro: knowsrec.intros elim: knowsrec.cases) 
apply (force intro: knowsrec.intros elim: knowsrec.cases) 
apply (induct_tac rL)
apply ((rule allI)+,rule impI,  rule conjI);
apply (force intro: knowsrec.intros elim: knowsrec.cases) 
apply (force intro: knowsrec.intros elim: knowsrec.cases) 
apply ((rule allI)+,rule impI,  rule conjI);
apply (force intro: knowsrec.intros elim: knowsrec.cases) 
apply (force intro: knowsrec.intros elim: knowsrec.cases) 
done;

(* FORMATTED 
 lemma decomp_knowsrec:
        (?r1.0, ?r2.0, ?rL @ ?r' # ?rL') \<in> knowsrec ?c \<Longrightarrow>
        (?r1.0, ?r', ?rL') \<in> knowsrec ?c \<and> (?r', ?r2.0, ?rL) \<in> knowsrec ?c *)


(* the other way : recomposing *)
lemma recomp_knowsrec[rule_format]: 
    " \<forall> r1 r2. ((r1,r',rL)\<in>knowsrec c \<and> (r',r2,rL')\<in>knowsrec c) \<longrightarrow>(r1,r2,rL@r'#rL')\<in>knowsrec c"
apply (induct_tac rL')
apply (induct_tac rL)
apply simp+;
apply ((rule allI)+,rule impI);
apply (force intro: knowsrec.intros elim: knowsrec.cases) 
apply (force intro: knowsrec.intros knowsrec_rev elim: knowsrec.cases knowsreccases_rev) 
apply (induct_tac rL)
apply ((rule allI)+,rule impI);
apply (force intro: knowsrec.intros elim: knowsrec.cases) 
apply ((rule allI)+,rule impI);
apply (simp+,erule conjE);
apply (force intro: knowsrec.intros elim: knowsrec.cases) 
done;

(* FORMATTED 
 lemma recomp_knowsrec:
        (?r1.0, ?r', ?rL) \<in> knowsrec ?c \<and> (?r', ?r2.0, ?rL') \<in> knowsrec ?c \<Longrightarrow>
        (?r1.0, ?r2.0, ?rL @ ?r' # ?rL') \<in> knowsrec ?c                  *)

lemma turn_list_knowsrec[rule_format]: 
    "( (r,r,rL@r'#rL')\<in>knowsrec c\<Longrightarrow>(r',r',rL'@r#rL)\<in>knowsrec c)";
apply (drule decomp_knowsrec);
apply (rule  recomp_knowsrec);
apply simp;
done;

lemma pre_nocycle_act_upd[rule_format]: 
  "\<forall> r r'. (((r,r',rL)\<in>(knowsrec (c(A\<mapsto>(empty,a))))\<and> RefAct A \<notin>(set rL\<union>{r})) \<longrightarrow> (r,r',rL) \<in>knowsrec c)"
apply (induct_tac rL);
apply clarify;
apply (erule knowsrec.cases);
apply simp+;
apply (unfold knows_def ActiveObject_def RequestQueue_def OneRequest_def);
apply simp+;
apply auto;
apply (case_tac "Aa=A");apply force;
apply simp+;
apply (subgoal_tac "(RefFut f, RefFut f', []) \<in> knowsrec c")
apply force;
apply (intro knowsrec.intros)
apply (force simp: knows_def OneRequest_def RequestQueue_def ActiveObject_def) 
apply (case_tac "Aa=A");apply force;
apply simp+;
apply (force simp: knows_def OneRequest_def RequestQueue_def ActiveObject_def intro: knowsrec.intros) 
apply (case_tac "Aa=A");apply force;
apply simp+;
apply (force simp: knows_def OneRequest_def RequestQueue_def ActiveObject_def intro: knowsrec.intros) 

apply (case_tac "Aa=A");apply force;
apply simp+;
apply (force simp: knows_def OneRequest_def RequestQueue_def ActiveObject_def intro: knowsrec.intros) 

apply (erule knowsrec.cases);
apply simp+;
apply (erule conjE)+;
apply (drule_tac x=r'a in spec);
apply (drule_tac x=r'' in spec);
apply simp+;
apply (subgoal_tac "(ra,r'a)\<in>knows c");
apply (force intro: knowsrec.intros elim: knowsrec.cases);
apply (drule not_sym)
apply (unfold knows_def ActiveObject_def RequestQueue_def OneRequest_def);
apply simp+;
apply safe;
apply auto;
apply (case_tac "Aa=A");apply force;apply force;
apply (case_tac "Aa=A");apply force;apply force;
apply (case_tac "Aa=A");apply force;apply force;
done;


lemma pre_nocycle_act_upd'[rule_format]: 
  "\<lbrakk>(r,r',rL)\<in>(knowsrec (c(A\<mapsto>(empty,a)))); RefAct A \<notin>(set rL\<union>{r})\<rbrakk> \<Longrightarrow> (r,r',rL) \<in>knowsrec c"
apply (rule pre_nocycle_act_upd)
apply auto;
done;

lemma pre_nocycle_fut_upd[rule_format]: 
  "\<forall> r r'. ((A\<in> dom c \<and>(r,r',rL)\<in>(knowsrec (c(A\<mapsto>((c.\<^sub>RA)(f\<mapsto>a),c.\<^sub>AA))))
   \<and> RefFut f \<notin>(set rL\<union>{r}))                 \<longrightarrow> (r,r',rL) \<in>knowsrec c)"
apply (induct_tac rL);
apply (rule allI)+;apply (rule impI);apply (erule conjE)+;
apply (erule knowsrec.cases);
apply simp+;
apply (unfold knows_def ActiveObject_def RequestQueue_def OneRequest_def);
apply simp+;
apply auto;
apply (case_tac "fa=f");
apply simp+;
apply (subgoal_tac "(RefFut fa, RefFut f', []) \<in> knowsrec c")
apply force;
apply (intro knowsrec.intros)
apply (unfold knows_def ActiveObject_def RequestQueue_def OneRequest_def);
apply simp+;
apply (rule_tac x=Aa in bexI);
apply (case_tac "Aa=A");apply force;apply force;
apply (case_tac "Aa=A");apply force;apply force;

apply (case_tac "fa=f");apply simp;
apply (subgoal_tac "(RefFut fa, RefAct B, []) \<in> knowsrec c")
apply force;
apply (intro knowsrec.intros)
apply (unfold knows_def ActiveObject_def RequestQueue_def OneRequest_def);
apply simp+;
apply (rule_tac x=Aa in bexI);
apply (case_tac "Aa=A");
apply force;apply force;
apply (case_tac "Aa=A");apply force;apply force;

apply (subgoal_tac "(RefAct Aa, RefFut f', []) \<in> knowsrec c")
apply simp;apply (intro knowsrec.intros)
apply (unfold knows_def ActiveObject_def RequestQueue_def OneRequest_def);
apply simp+;
apply (case_tac "Aa=A");apply force;apply force;

apply (intro knowsrec.intros)
apply (unfold knows_def ActiveObject_def RequestQueue_def OneRequest_def);
apply simp+;
apply (case_tac "Aa=A");apply force;apply force;

apply (erule knowsrec.cases);
apply simp+;
apply (erule conjE)+;
apply (drule_tac x=r'a in spec);
apply (drule_tac x=r'' in spec);
apply simp+;
apply (subgoal_tac "(ra,r'a)\<in>knows c");
apply (force intro: knowsrec.intros elim: knowsrec.cases);
apply (drule not_sym)
apply (drule not_sym)
apply (unfold knows_def ActiveObject_def RequestQueue_def OneRequest_def);
apply simp+;
apply safe;
apply auto;
apply (case_tac "Aa=A");apply force;apply force;
apply (case_tac "Aa=A");apply simp+;
apply (erule conjE);
apply (rotate_tac -2,drule sym,simp+);
apply (case_tac "fa=f");apply force;apply force;
apply (case_tac "fa=f");apply force;apply force;
apply (case_tac "Aa=A");apply force;apply force;
apply (case_tac "Aa=A");apply force;apply force;
apply (case_tac "Aa=A");apply force;apply force;
apply (case_tac "Aa=A");apply force;apply force;
done;

lemma pre_nocycle_fut_upd'[rule_format]: 
  "\<lbrakk>(r,r',rL)\<in>(knowsrec (c(A\<mapsto>((c.\<^sub>RA)(f\<mapsto>a),c.\<^sub>AA))));A\<in> dom c ;RefFut f \<notin>(set rL\<union>{r})\<rbrakk>                              \<Longrightarrow>(r,r',rL) \<in>knowsrec c"
apply (rule pre_nocycle_fut_upd)
apply auto;
done;


lemma nocycle_act_upd: 
"\<lbrakk>nocycle c; (RefAct A,RefAct A)\<notin>((knows (c(A\<mapsto>(empty,a))))\<^sup>+)\<rbrakk>\<Longrightarrow>nocycle (c(A\<mapsto>(empty,a)))"
apply (unfold nocycle_def);
apply (simp add: knowsrecI);
apply (rule allI)+;
apply (case_tac "r=RefAct A");
apply force;
apply (case_tac "RefAct A\<in>set rL")
apply (drule list_set_decomp);
apply (erule exE)+;
apply (rule notI)
apply simp;
apply (drule turn_list_knowsrec);
apply force;
apply (rule notI)
apply (drule pre_nocycle_act_upd');
apply auto;
done;

lemma nocycle_fut_upd: 
"\<lbrakk>nocycle c;A\<in> dom c; (RefFut f,RefFut f)\<notin>((knows (c(A\<mapsto>((c.\<^sub>RA)(f\<mapsto>a),c.\<^sub>AA))))\<^sup>+)\<rbrakk>\<Longrightarrow>nocycle (c(A\<mapsto>((c.\<^sub>RA)(f\<mapsto>a),c.\<^sub>AA)))"
apply (unfold nocycle_def);
apply (simp add: knowsrecI);
apply (rule allI)+;
apply (case_tac "r=RefFut f");
apply force;
apply (case_tac "RefFut f\<in>set rL")
apply (drule list_set_decomp);
apply (erule exE)+;
apply (rule notI)
apply simp;
apply (drule turn_list_knowsrec);
apply force;
apply (rule notI)
apply (drule pre_nocycle_fut_upd');
apply auto;
done;

(* minimality argument on paths: if there is a path between A and B then there is one that 
   does not pass by A and B *)
lemma minimal_path[rule_format]: " ((r,r',rL0)\<in>(knowsrec c)\<longrightarrow> (\<exists>rL. ((r,r',rL)\<in> knowsrec c\<and>r\<notin>set rL\<and>r'\<notin>set rL)))"
apply (rule length_induct)
apply auto;
apply (case_tac "r\<in>set xs")
apply (drule list_set_decomp)
apply (erule exE)+;
apply simp+;
apply (drule decomp_knowsrec)
apply auto;
apply (case_tac "r'\<in>set xs")
apply (drule list_set_decomp)
apply (erule exE)+;
apply simp+;
apply (drule decomp_knowsrec)
apply auto;
done;


lemma add_act_inc: "G\<notin>dom c \<Longrightarrow>(knows c)\<^sup>+\<subseteq>(knows (c(G\<mapsto>(empty,a))))\<^sup>+"
apply (rule subtrancl);
apply (force simp: knows_def RequestQueue_def ActiveObject_def OneRequest_def)
done;

lemma act_upd_knows: "\<lbrakk>(r,r')\<in>knows c;G\<notin>dom c\<rbrakk>\<Longrightarrow>(r,r')\<in>knows (c(G\<mapsto>(empty,a)))"
apply (auto simp: knows_def ActiveObject_def RequestQueue_def OneRequest_def)
apply (case_tac "A=G");apply force;apply force;
apply (case_tac "A=G");apply force;apply force;
apply (case_tac "A=G");apply force;apply force;
apply (case_tac "A=G");apply force;apply force;
done;

lemma not_indom_not_in_knows: "c G = None \<Longrightarrow> (RefAct G,r)\<notin>knows c"
by (auto simp: knows_def RequestQueue_def OneRequest_def ActiveObject_def);

lemma wf_knows_act: "(RefAct A,r)\<in>knows c \<Longrightarrow>A \<in> dom c"
by (auto simp: knows_def);

lemma wf_knowsactrec: "(RefAct A,r,rL)\<in>knowsrec c \<Longrightarrow>A \<in> dom c"
apply (erule knowsrec.cases)
apply (auto simp: wf_knows_act);
apply (drule wf_knows_act);
apply force;
done;

(* wf lemmas sate relate references existing in the knows relation and in the configurtions *)
lemma wf_knows_fut: "(RefFut f,r)\<in>knows c \<Longrightarrow>\<exists> A \<in> dom c. f \<in> dom (c.\<^sub>R A)"
by (auto simp: knows_def RequestQueue_def);

lemma wf_knowsfutrec: "(RefFut f,r,rL)\<in>knowsrec c \<Longrightarrow>\<exists> A \<in> dom c. f \<in> dom (c.\<^sub>R A)"
apply (erule knowsrec.cases)
apply (auto simp: wf_knows_fut);
done;

lemma rev_list_notNil[rule_format]: "xs\<noteq>[]\<longrightarrow>(\<exists>x ys. xs = ys @ [x])"
apply (induct_tac xs);
apply auto;
done;

lemma wellformed_knows_act: "\<lbrakk>(r,RefAct A)\<in>knows c;wf_conf c \<rbrakk> \<Longrightarrow>A \<in> dom c"
apply (unfold knows_def)
apply auto;
apply (simp add: wf_conf_def wf_1b_def);
apply (erule conjE)+; 
apply (simp only:  OneRequest_def RequestQueue_def);
apply (drule_tac x=Aa in bspec);
apply force;
apply (drule_tac x=f in bspec)
apply force;apply force;
apply (simp add: wf_conf_def wf_1a_def);
apply (erule conjE)+; 
apply (simp only:  ActiveObject_def);
apply (drule_tac x=Aa in bspec);
apply force;
apply force;
done;

lemma wellformed_knowsactrec: "\<lbrakk>(r,RefAct A,rL)\<in>knowsrec c;wf_conf c \<rbrakk> \<Longrightarrow>A \<in> dom c"
apply (case_tac "rL=[]")
apply (erule knowsrec.cases)
apply (auto simp: wellformed_knows_act);
apply (drule rev_list_notNil);
apply (erule exE)+;
apply (simp,drule knowsreccases_rev);
apply (erule conjE);
apply (drule wellformed_knows_act);
apply (auto);
done;

lemma wellformed_knows_fut: "\<lbrakk>(r,RefFut f)\<in>knows c;wf_conf c \<rbrakk> \<Longrightarrow>\<exists> A \<in> dom c. f \<in> dom (c.\<^sub>R A)"
apply (unfold knows_def)
apply auto;
apply (simp add: wf_conf_def wf_2b_def);
apply (erule conjE)+; 
apply (simp only:  OneRequest_def RequestQueue_def);
apply (drule_tac x=A in bspec);
apply force;
apply (drule_tac x=fa in bspec)
apply force;apply force;
apply (simp add: wf_conf_def wf_2a_def);
apply (erule conjE)+; 
apply (simp only:  ActiveObject_def);
apply (drule_tac x=A in bspec);
apply force;
apply force;
done;

lemma wellformed_knowsfutrec: "\<lbrakk>(r,RefFut f,rL)\<in>knowsrec c;wf_conf c \<rbrakk> \<Longrightarrow>\<exists> A \<in> dom c. f \<in> dom (c.\<^sub>R A)"
apply (case_tac "rL=[]")
apply (erule knowsrec.cases)
apply (auto simp: wellformed_knows_fut RequestQueue_def);
apply (drule wellformed_knows_fut);
apply (auto simp: RequestQueue_def);
apply (drule rev_list_notNil);
apply (erule exE)+;
apply (simp,drule knowsreccases_rev);
apply (erule conjE);
apply (drule wellformed_knows_fut);
apply (auto simp: RequestQueue_def);
done;

(* the main theorem: reduction maintains absence of cycles, note that wellformedness is required here *)
theorem nocycle_SR: "\<lbrakk>c \<rightarrow>\<^sub>\<zeta> c';nocycle c;disjunct_futures c; wf_conf c\<rbrakk> \<Longrightarrow> nocycle c'"

 apply (frule_tac P=" nocycle c'" in zeta.cases);

(* LOCAL *)
 apply simp;
 apply (subgoal_tac "(c(A \<mapsto> (m(fk \<mapsto> a'), a'')))=(c(A\<mapsto>((c.\<^sub>RA)(fk\<mapsto>a'),c.\<^sub>AA)))"); (* rewriting  *)
   (* 8 *)
 apply simp;
 apply (rule_tac nocycle_fut_upd)
 apply simp+;
 apply force;
 apply (rule notI);
 apply (simp add: knowsrecI);
 apply (erule exE,drule minimal_path);
 apply (erule exE,(erule conjE)+);
 
 (* Main subgoal : *)
 apply (subgoal_tac 
   " \<forall> r.(RefFut fk, r)\<in> knows (ca(A \<mapsto> ((ca .\<^sub>R A)(fk \<mapsto> a'), ca .\<^sub>A A)))
                                            \<longrightarrow> (RefFut fk, r)\<in> knows c")
 apply (erule knowsrec.cases);
 (* direct *)
 apply simp+;
 apply (drule_tac x=r in spec)
 apply simp+;
 apply (simp add: nocycle_def)
 apply blast;
  (* rec *)
 apply simp+;
 apply (erule conjE)+;
 apply (drule_tac x=r' in spec)
 apply (rotate_tac -7,drule sym);
  apply simp+;
 apply (drule pre_nocycle_fut_upd')
 apply blast;
 apply blast;
 apply (subgoal_tac "(RefFut fk, RefFut fk,r' # rLb) \<in> knowsrec c");
  apply (simp add: nocycle_def knowsrecI)
 apply (force intro: knowsrec.intros)
 (* the main subgoal 8 *)
 apply (rule allI);
 apply (rule impI);
 apply (thin_tac "wf_conf ca");
 apply (thin_tac " c' = ca(A \<mapsto> ((ca .\<^sub>R A)(fk \<mapsto> a'), ca .\<^sub>A A))");
 apply (thin_tac "  ca(A \<mapsto> (m(fk \<mapsto> a'), a'')) = ca(A \<mapsto> ((ca .\<^sub>R A)(fk \<mapsto> a'), ca .\<^sub>A A))");
 apply simp+;
 apply (unfold knows_def  OneRequest_def RequestQueue_def ActiveObject_def);
 apply simp;
 apply auto;
 apply (drule_tac E=E and fk=f' in  local_and_futref)
 apply simp+;
 apply (erule exE);
 apply (rule_tac x=A in bexI)
 apply force;apply force;
 apply (subgoal_tac "Aa=A");
 apply simp+;

 (* again *)  apply (drule_tac E=E and fk=f' in  local_and_futref);apply simp+; apply (erule exE); apply (rule_tac x=A in bexI); apply force;apply force;
 
 apply (simp add: disjunct_futures_def)
 apply (drule_tac x=A in bspec);apply force;
 apply (drule_tac x=Aa in bspec);apply force;
 apply (force simp: RequestQueue_def)

 apply (drule_tac E=E and G=B in  local_and_actref)
 apply (erule exE);
 apply force;

 apply (subgoal_tac "Aa=A");
 apply simp+;

 (* again *)   apply (drule_tac E=E and G=B in  local_and_actref);apply (erule exE);  apply force;
 apply (simp add: disjunct_futures_def)
 apply (drule_tac x=A in bspec);apply force;
 apply (drule_tac x=Aa in bspec);apply force;
 apply (force simp: RequestQueue_def)

(* NEWACT: activity creation 5 *)
 apply (subgoal_tac  "(RefFut fk, RefFut fk) \<notin> (knows (c(A \<mapsto> (m(fk \<mapsto> E \<up> ActRef G), a''), G \<mapsto> (empty, a))))");
 prefer 2;
 apply (simp add: knows_def ActiveObject_def RequestQueue_def OneRequest_def);
 apply auto;
 apply (subgoal_tac " \<exists> (E0::general_context).  E \<up> (Active  a) = E0 \<up> FutRef fk")
 apply (erule exE)
 apply (subgoal_tac "(RefFut fk, RefFut fk) \<in> knows c")
 apply (force simp: nocycle_def knowsrecI intro: knowsrec.intros)
 apply (simp add: knows_def ActiveObject_def RequestQueue_def OneRequest_def);
 apply (rule_tac x=A in bexI);
 apply simp+;
 apply (rule conjI,force);
 apply (rule_tac x=E0 in exI);
 apply force;apply force;
 apply (drule context_inc_Fut);
 apply (force simp: diff_futactrefs);

 (* again :-( *)
 apply (subgoal_tac " \<exists> (E0::general_context).  E \<up> (Active  a) = E0 \<up> FutRef fk")
 apply (erule exE)
 apply (subgoal_tac "(RefFut fk, RefFut fk) \<in> knows c")
 apply (force simp: nocycle_def knowsrecI intro: knowsrec.intros)
 apply (simp add: knows_def ActiveObject_def RequestQueue_def OneRequest_def);
 apply (rule_tac x=A in bexI);
 apply simp+;
 apply (rule conjI,force);
 apply (rule_tac x=E0 in exI);
 apply force;apply force;
 apply (drule context_inc_Fut);
 apply (force simp: diff_futactrefs);

 apply (subgoal_tac "(RefFut fk, RefFut fk) \<in> knows c")
 apply (force simp: nocycle_def knowsrecI intro: knowsrec.intros)
 apply (simp add: knows_def ActiveObject_def RequestQueue_def OneRequest_def);
 apply (rule_tac x=Aa in bexI);
 apply simp+;
 apply (rule conjI,force);
 apply (rule_tac x=Ea in exI);
 apply force;apply force;


 apply (subgoal_tac"(RefAct G,RefFut fk) \<notin> (knows (c(A \<mapsto> (m(fk \<mapsto> E \<up> ActRef G), a''),G \<mapsto> (empty, a))))\<^sup>+" )
 prefer 2;
 apply (subgoal_tac "(RefAct G,RefFut fk) \<in> (knows (c(A \<mapsto> (m(fk \<mapsto> E \<up> ActRef G), a''),G \<mapsto> (empty, a))))\<^sup>+\<longrightarrow>(RefFut fk,RefFut fk) \<in> (knows c)\<^sup>+")
 apply (force simp: nocycle_def);
 apply (rule impI);
 apply (simp add: knowsrecI);
 apply (erule exE);
 apply (drule minimal_path,erule exE,erule conjE)
 apply simp+;

 (* 6 *)
 apply (subgoal_tac 
       "\<forall> r'.((RefAct G, r')\<in> knows (c(A \<mapsto> (m(fk \<mapsto> E \<up> ActRef G), a''), G \<mapsto> (empty, a)))
                           \<longrightarrow>(RefFut fk, r')\<in> knows c)")
 prefer 2;
 apply (simp add: knows_def ActiveObject_def RequestQueue_def OneRequest_def);
 apply (subgoal_tac "A\<noteq>G")
 prefer 2;apply force;
 apply auto;
 apply (rule_tac x=A in bexI); apply simp+;
 apply (rule conjI,force);
 apply (rule_tac x="E \<circ>(cActive (Ea  \<circ> cHole))" in exI);
 apply force;apply force;
 apply (rotate_tac -1, drule_tac x=A in bspec);
 apply force;apply simp+;
 apply (rotate_tac -1, erule impE);apply force;
 apply (rotate_tac -1);
 apply (drule_tac x="E \<circ>(cActive (Ea  \<circ> cHole))" in spec);
 apply simp+;

 (* 6 again *)
 apply (erule knowsrec.cases);
 apply simp;
 apply (rotate_tac -4,drule sym,drule sym);
 apply simp+;
 apply (drule_tac x="RefFut fk" in spec);
 apply (force intro: knowsrec.intros);

 apply simp;
 apply (erule conjE)+;
 apply (rotate_tac -8,drule sym);
 apply (rotate_tac -2,drule sym);
 apply simp+;
 apply (drule_tac x=r' in spec);
 apply simp+;
 apply (rule_tac x="r'#rL" in exI);
 apply (subgoal_tac "(r', RefFut fk, rL) \<in> knowsrec c");
 apply (force intro: knowsrec.intros);
 apply (drule_tac A=G and a=a in pre_nocycle_act_upd')
 apply force;
 apply (rule_tac A=A and f=fk and a=" E \<up> ActRef G" in pre_nocycle_fut_upd')
 apply (force simp: ActiveObject_def RequestQueue_def);
 apply force;apply force;
 apply (rule nocycle_act_upd);
 apply (subgoal_tac "(c(A \<mapsto> (m(fk \<mapsto> E \<up> ActRef G), a'')))=(c(A \<mapsto> ((c .\<^sub>R A)(fk \<mapsto> E \<up> ActRef G), c .\<^sub>A A)))");
 apply (simp,rule nocycle_fut_upd);
 apply simp+;apply force;
 apply (simp add: knowsrecI)
 apply (rule allI,rule notI);
 apply (case_tac "RefAct G \<in>set rL")
 apply (drule list_set_decomp,(erule exE)+,simp);
 apply (drule turn_list_knowsrec);
 apply (rotate_tac -1,erule knowsrec.cases);
 apply (force simp: ActiveObject_def knows_def RequestQueue_def);
 apply simp+;
 apply (erule conjE)+;
 apply (rotate_tac -4,drule sym);
 apply simp+;
 apply (subgoal_tac "(c(A \<mapsto> ((c .\<^sub>R A)(fk \<mapsto> E \<up> ActRef G), c .\<^sub>A A))) G = None",rotate_tac -1)
 apply (drule_tac r=r' in not_indom_not_in_knows);

 (* eliminate a few hypotheses ... *)
 apply (thin_tac "c(A \<mapsto> (m(fk \<mapsto> E \<up> ActRef G), a'')) = c(A \<mapsto>
        ((c .\<^sub>R A)(fk \<mapsto> E \<up> ActRef G), c .\<^sub>A A))")
 apply simp;
 apply (simp add: ActiveObject_def knows_def RequestQueue_def OneRequest_def);
 apply force;
 apply force;
 apply (drule minimal_path,erule exE,(erule conjE)+);
 apply (erule knowsrec.cases)
 apply (thin_tac "c(A \<mapsto> (m(fk \<mapsto> E \<up> ActRef G), a'')) = c(A \<mapsto>
        ((c .\<^sub>R A)(fk \<mapsto> E \<up> ActRef G), c .\<^sub>A A))")
 apply (rotate_tac 7);
 apply (erule rev_notE);
 apply simp+;
 apply (rule_tac G=G in act_upd_knows)
 apply force;apply simp+;
 apply (force);
 apply simp+;apply (erule conjE)+;
 apply (rotate_tac -6,drule sym);
 apply simp+;
 apply (drule_tac A=A and f=fk and a=" E \<up> ActRef G" in pre_nocycle_fut_upd')
 apply force;apply force;
 apply (subgoal_tac "(RefFut fk, r') \<in> knows c");
 apply (force intro: knowsrec.intros simp: nocycle_def knowsrecI);
 
(* eliminate a few hypotheses ... *)
 apply (rotate_tac 12);
 apply (drule_tac Q= "c G = None" in conjI,assumption,drule conjunct2);
 apply (thin_tac "c(A \<mapsto> (m(fk \<mapsto> E \<up> ActRef G), a'')) = c(A \<mapsto>
        ((c .\<^sub>R A)(fk \<mapsto> E \<up> ActRef G), c .\<^sub>A A))");
 apply (simp add: knows_def ActiveObject_def RequestQueue_def OneRequest_def);
 apply auto;
 apply (rotate_tac -3)
 apply (drule_tac x=A in bspec);
 apply force;
 apply simp+;
 apply (erule impE);
 apply force;
 apply (drule context_inc_Fut);
 apply (erule disjE,erule exE);
 apply (force simp: diff_futactrefs)
 apply (rotate_tac -1, drule_tac x="Active a" in spec);
 apply force;
 apply (rotate_tac -4)
 apply (drule_tac x=A in bspec);
 apply force;
 apply simp+;apply (erule impE);apply force;
 apply (simp add: disjunct_futures_def)
 apply (rotate_tac 7)
 apply (drule_tac x=A in bspec);apply force;
 apply (rotate_tac -1)
 apply (drule_tac x=Aa in bspec);apply force;
 apply (case_tac "Aa=A");apply simp+;
 apply (drule context_inc_Fut);
 apply (erule disjE,erule exE);
 apply (force simp: diff_futactrefs)
 apply (rotate_tac -1, drule_tac x="Active a" in spec);
 apply force;
 apply simp+;
 apply (subgoal_tac "fk\<notin>dom (c .\<^sub>R Aa)")
 apply (force simp: RequestQueue_def)
 apply (force simp: RequestQueue_def)

 apply (drule context_inc_Act);
 apply (erule disjE,erule exE);
 apply (subgoal_tac "G\<noteq>B");
 apply (drule diff_actrefs);
 apply force;
 apply (subgoal_tac "B\<in>dom c");
 apply force;
 apply (rule wf_knowsactrec);
 apply simp;
 apply (drule_tac x="Active a" in spec);
 apply (erule exE);
 apply simp+;
 apply (rotate_tac -3)
 apply (drule_tac x=A in bspec)
 apply force;
 apply (subgoal_tac "A\<noteq>G");
 apply simp+;
 apply force;
 apply force;
 apply (subgoal_tac "A\<noteq>G");
 apply (case_tac "Aa=A");apply simp+;

 apply (drule context_inc_Act);
 apply (erule disjE,erule exE);
 apply (subgoal_tac "G\<noteq>B")
 apply (rotate_tac -1)
 apply (drule diff_actrefs);
 apply force;
 apply (subgoal_tac "B\<in>dom c");
 apply force;
 apply (rule wf_knowsactrec);
 apply simp;
 apply (drule_tac x="Active a" in spec);
 apply (erule exE);
 apply simp+;
 apply (rotate_tac -6)
 apply (drule_tac x=A in bspec)
 apply force;
 apply (subgoal_tac "A\<noteq>G");
 apply simp+;
 apply force;
 apply force;
 apply simp+;
 apply (simp add: disjunct_futures_def)
 apply (rotate_tac 4)
 apply (drule_tac x=A in bspec);apply force;
 apply (rotate_tac -1)
 apply (drule_tac x=Aa in bspec);apply force;
 apply (subgoal_tac "fk\<notin>dom (c .\<^sub>R Aa)")
 apply (force simp: RequestQueue_def)
 apply (force simp: RequestQueue_def)
 apply force;
 apply (force simp: RequestQueue_def ActiveObject_def)

 (* no cycle with G *)
 apply (simp add: knowsrecI)
 apply (erule exE);
 apply (case_tac "rL=[]")
 apply simp+;apply (erule knowsrec.cases)
 apply simp+;
 apply (rotate_tac -2,drule sym);
 apply simp+;
 apply (subgoal_tac "\<exists> (E::general_context). a=E\<up>ActRef G");
 apply (erule exE);apply simp+;
 apply (simp add: wf_conf_def wf_1b_def);apply (erule conjE)+;
 apply (rotate_tac -1, drule_tac x=A in bspec);apply force;
 apply (rotate_tac -1, drule_tac x=fk in bspec);apply (force simp: RequestQueue_def);
 apply (rotate_tac -1, drule_tac x="E\<circ> (cActive (Ea))" in spec);
 apply (rotate_tac -1, drule_tac x=G in spec);
 apply (force simp: OneRequest_def);
 apply (force simp: knows_def RequestQueue_def OneRequest_def ActiveObject_def)
 apply force;
 apply (drule rev_list_notNil)
 apply (erule exE)+;
 apply simp+;
 apply (drule knowsreccases_rev)
 apply (erule conjE);
 apply (subgoal_tac "x=RefFut fk")
 apply simp+;
 apply (case_tac x);prefer 2;
 apply simp+;
 apply (simp add: wf_conf_def wf_1b_def);apply (erule conjE)+;
 apply (rotate_tac -3,erule contrapos_pp)
 apply (simp add: knows_def RequestQueue_def OneRequest_def)
 apply (erule disjE);
 apply (rule_tac x=A in bexI);
 apply (erule conjE,erule exE);
 apply (rule_tac x=FutureRef in bexI);
 apply (rule_tac x=Ea in exI);
 apply (rule_tac x=G in exI);
 apply simp+;
 apply (subgoal_tac "A\<noteq>G");
 apply simp+;
 apply force;
 apply force;
 apply (subgoal_tac "A\<noteq>G");
 apply simp+;
 apply force;
 apply force;
 apply (erule bexE);
 apply (erule conjE);
 apply (subgoal_tac "A\<noteq>G");
 apply (subgoal_tac "Aa\<noteq>G");
 apply (rule_tac x=Aa in bexI);
 apply (erule conjE,erule exE);
 apply (rule_tac x=FutureRef in bexI);
 apply (rule_tac x=Ea in exI);
 apply (rule_tac x=G in exI);
 apply simp+;apply auto;
 apply (case_tac "Aa=A");apply force;apply simp+;
 apply (case_tac "Aa=A");  apply force; apply force;

 apply (case_tac "ActivityRef=G");
 apply simp+;
 apply (simp add: knows_def RequestQueue_def OneRequest_def ActiveObject_def)
 apply auto;
 apply (simp add: wf_conf_def wf_1b_def);apply (erule conjE)+;
 apply (rotate_tac -3,erule contrapos_pp)
 apply simp+;
 apply (rule_tac x=A in bexI);
 apply (rule_tac x=fk in bexI);
 apply (rule_tac x="(E \<circ>  (cActive Ea))" in exI);
 apply (rule_tac x=G in exI);
 apply (simp add: OneRequest_def)
 apply force;
 apply (force simp: RequestQueue_def)
 apply force;
 
 apply (simp add: wf_conf_def wf_1a_def);apply (erule conjE)+;
 apply (rotate_tac -4,erule contrapos_pp)
 apply (simp add: knows_def RequestQueue_def OneRequest_def ActiveObject_def)
 apply (subgoal_tac "A\<noteq>G");
 apply (subgoal_tac "ActivityRef\<noteq>G");
 apply (rule_tac x=ActivityRef in bexI);
 apply (erule conjE,erule exE);
 apply (rule_tac x=Ea in exI);
 apply (rule_tac x=G in exI);
 apply simp+;
 apply (case_tac "ActivityRef=A");apply force;apply force;
 apply force;
 apply (case_tac "ActivityRef=A");apply force;apply force;
 apply force;
  
 prefer 3;
 (* UPDATE AO *)

 apply (subgoal_tac  "(RefFut fk, RefFut fk) \<notin> (knows (c(A' \<mapsto> (m'(fk \<mapsto> E \<up> ActRef G), a'), G \<mapsto> (empty, Upd a li b))))");
 prefer 2;
 apply (simp add: knows_def ActiveObject_def RequestQueue_def OneRequest_def);
 apply auto;
 apply (subgoal_tac " \<exists> (E0::general_context).  E \<up> ( Upd (ActRef B) li b) = E0 \<up> FutRef fk")
 apply (erule exE)
 apply (subgoal_tac "(RefFut fk, RefFut fk) \<in> knows c")
 apply (force simp: nocycle_def knowsrecI intro: knowsrec.intros)
 apply (simp add: knows_def ActiveObject_def RequestQueue_def OneRequest_def);
 apply (rule_tac x=A' in bexI);
 apply simp+;
 apply (rule conjI,force);
 apply (rule_tac x=E0 in exI);
 apply force;apply force;
 apply (drule context_inc_Fut);
 apply (force simp: diff_futactrefs);

 (* again :-( 6 *)
 apply (subgoal_tac " \<exists> (E0::general_context).  E \<up> ( Upd (ActRef B) li b) = E0 \<up> FutRef fk")
 apply (erule exE)
 apply (subgoal_tac "(RefFut fk, RefFut fk) \<in> knows c")
 apply (force simp: nocycle_def knowsrecI intro: knowsrec.intros)
 apply (simp add: knows_def ActiveObject_def RequestQueue_def OneRequest_def);
 apply (rule_tac x=A' in bexI);
 apply simp+;
 apply (rule conjI,force);
 apply (rule_tac x=E0 in exI);
 apply force;apply force;
 apply (drule context_inc_Fut);
 apply (force simp: diff_futactrefs);

 apply (subgoal_tac "(RefFut fk, RefFut fk) \<in> knows c")
 apply (force simp: nocycle_def knowsrecI intro: knowsrec.intros)
 apply (simp add: knows_def ActiveObject_def RequestQueue_def OneRequest_def);
 apply (rule_tac x=A in bexI);
 apply simp+;
 apply (rule conjI,force);
 apply (rule_tac x=Ea in exI);
 apply force;apply force;

 (* a crucial subgoal *)
 (* 4 *)
 apply (subgoal_tac 
       "\<forall> r'.((RefAct G, r')\<in> knows (c(A' \<mapsto> (m'(fk \<mapsto> E \<up> ActRef G), a'), G \<mapsto> (empty, Upd a li b)))
                           \<longrightarrow>((RefFut fk, r')\<in> knows c\<or>(RefAct B, r')\<in> knows c))")
 prefer 2;
 apply (simp add: knows_def ActiveObject_def RequestQueue_def OneRequest_def);
 apply (subgoal_tac "B\<noteq>G")
 prefer 2;apply force;
 apply (subgoal_tac "A'\<noteq>G")
 prefer 2;apply force;
 apply auto;

 apply (rotate_tac -2,drule sym,frule context_update);
 apply (erule disjE)+;
 apply simp+;apply (erule exE,simp);
 apply (drule_tac x=A' in bspec); apply force;
 apply (erule impE);apply force;
 apply (rotate_tac -1); apply (drule_tac x="E \<circ>(cUpdR (ActRef B) li E')" in spec);
 apply force;
  apply (rotate_tac -2, drule_tac x=A' in bspec);
 apply force;apply simp+;
 apply (rotate_tac -1, erule impE);apply force;
 apply (rotate_tac -1);
 apply (rotate_tac -1,drule sym,frule context_update);
 apply (erule disjE)+;
 apply simp+;apply (erule exE,simp);
 apply (drule_tac x="E \<circ>(cUpdR (ActRef B) li E')" in spec);
 apply simp+;


 (* 4 *)
 apply (subgoal_tac"(RefAct G,RefFut fk) \<notin> (knows (c(A' \<mapsto> (m'(fk \<mapsto> E \<up> ActRef G), a'),G \<mapsto> (empty, Upd a li b))))\<^sup>+" )
 prefer 2;
 apply (subgoal_tac "(RefAct G,RefFut fk) \<in> (knows (c(A' \<mapsto> (m'(fk \<mapsto> E \<up> ActRef G), a'),G \<mapsto> (empty, Upd a li b))))\<^sup>+\<longrightarrow>(RefFut fk,RefFut fk) \<in> (knows c)\<^sup>+")
 apply (force simp: nocycle_def);
 
 (* we use the crucial subgoal  here! 5 *)
 apply (rule impI);
 apply (simp add: knowsrecI);
 apply (erule exE);
 apply (drule minimal_path,erule exE,erule conjE)
 apply simp+;

 apply (erule knowsrec.cases);
 apply simp;
 apply (rotate_tac -4,drule sym,drule sym);
 apply simp+;
 apply (drule_tac x="RefFut fk" in spec);
 apply simp+;
 apply (erule disjE)
 apply (force intro: knowsrec.intros);
 apply (subgoal_tac "(RefFut fk,RefAct B) \<in> knows c")
 apply (force intro: knowsrec.intros);
 apply (simp add: knows_def RequestQueue_def OneRequest_def)
 apply (rule_tac x=A' in bexI)
 apply (rule conjI);apply force;
 apply (rule_tac x="E \<circ> (cUpdL cHole li b)" in exI)
 apply force;apply force;

 (* 5 again *)
 apply simp;
 apply (erule conjE)+;
 apply (rotate_tac -8,drule sym);
 apply (rotate_tac -2,drule sym);
 apply simp+;
 apply (drule_tac x=r' in spec);
 apply simp+;
 apply (subgoal_tac "(r', RefFut fk, rL) \<in> knowsrec c");
 apply (erule disjE)
 apply (rule_tac x="r'#rL" in exI);
 apply (force intro: knowsrec.intros);
 apply (subgoal_tac "(RefFut fk,RefAct B) \<in> knows c")
 apply (rule_tac x="(RefAct B)#r'#rL" in exI);
 apply (force intro: knowsrec.intros);
 apply (simp add: knows_def RequestQueue_def OneRequest_def)
 apply (rule_tac x=A' in bexI)
 apply (rule conjI);apply force;
 apply (rule_tac x="E \<circ> (cUpdL cHole li b)" in exI)
 apply force;apply force;
 apply (drule_tac A=G and a="Upd a li b" in pre_nocycle_act_upd')
 apply force;
 apply (rule_tac A=A' and f=fk and a=" E \<up> ActRef G" in pre_nocycle_fut_upd')
 apply (force simp: ActiveObject_def RequestQueue_def);
 apply force;apply force;

 (* 4 :-) *)
 apply (rule nocycle_act_upd);
 apply (subgoal_tac "(c(A' \<mapsto> (m'(fk \<mapsto> E \<up> ActRef G), a')))=(c(A' \<mapsto> ((c .\<^sub>R A')(fk \<mapsto> E \<up> ActRef G), c .\<^sub>A A')))");
 apply (simp,rule nocycle_fut_upd);
 apply simp+;apply force;
 apply (simp add: knowsrecI)
 apply (rule allI,rule notI);
 apply (case_tac "RefAct G \<in>set rL")
 apply (drule list_set_decomp,(erule exE)+,simp);
 apply (drule turn_list_knowsrec);
 apply (rotate_tac -1,erule knowsrec.cases);
 apply (force simp: ActiveObject_def knows_def RequestQueue_def);
 apply simp+;
 apply (erule conjE)+;
 apply (rotate_tac -4,drule sym);
 apply simp+;
 apply (subgoal_tac "(c(A' \<mapsto> ((c .\<^sub>R A')(fk \<mapsto> E \<up> ActRef G), c .\<^sub>A A'))) G = None",rotate_tac -1)
 apply (drule_tac r=r' in not_indom_not_in_knows);

 (* eliminate a few hypotheses ... *)
 apply (rotate_tac 12);
 apply (drule_tac Q= "rL2 @ RefFut fk # rL1 = r' # rLa" in conjI,assumption,drule conjunct2);
 apply (drule_tac Q= "rL2 @ RefFut fk # rL1 = r' # rLa" in conjI,assumption,drule conjunct2);
 apply (thin_tac "c(A' \<mapsto> (m'(fk \<mapsto> E \<up> ActRef G), a')) = c(A' \<mapsto>
        ((c .\<^sub>R A')(fk \<mapsto> E \<up> ActRef G), c .\<^sub>A A'))");
 apply simp;
 apply (simp add: ActiveObject_def knows_def RequestQueue_def OneRequest_def);
 apply force;
 apply force;
 apply (drule minimal_path,erule exE,(erule conjE)+);
 apply (erule knowsrec.cases)
 apply (rotate_tac 8);
 apply (erule rev_notE);
 apply simp;
 apply (rule_tac G=G in act_upd_knows)
 apply force;apply simp+;
 apply (force);
 apply simp+;apply (erule conjE)+;
 apply (rotate_tac -6,drule sym);
 apply simp+;
 apply (drule_tac A=A' and f=fk and a=" E \<up> ActRef G" in pre_nocycle_fut_upd')
 apply force;apply force;
 apply (subgoal_tac "(RefFut fk, r') \<in> knows c");
 apply (force intro: knowsrec.intros simp: nocycle_def knowsrecI);
 
(* eliminate a few hypotheses ... *)
 apply (rotate_tac 13);
 apply (drule_tac Q= "c G = None" in conjI,assumption,drule conjunct2);
 apply (drule_tac Q= "c G = None" in conjI,assumption,drule conjunct2);
 apply (thin_tac "c(A' \<mapsto> (m'(fk \<mapsto> E \<up> ActRef G), a')) = c(A' \<mapsto>
        ((c .\<^sub>R A')(fk \<mapsto> E \<up> ActRef G), c .\<^sub>A A'))");
 apply (simp add: knows_def ActiveObject_def RequestQueue_def OneRequest_def);
 apply auto;
 apply (rotate_tac -3)
 apply (drule_tac x=A' in bspec);
 apply force;
 apply simp+;apply (erule impE);apply force;
 apply (drule context_inc_Fut);
 apply (erule disjE,erule exE);
 apply (force simp: diff_futactrefs)
 apply (rotate_tac -1, drule_tac x="Upd (ActRef B) li b" in spec);
 apply force;
 apply (rotate_tac -4)
 apply (drule_tac x=A' in bspec);
 apply force;
 apply simp+;apply (erule impE);apply force;
 apply (simp add: disjunct_futures_def)
 apply (rotate_tac 7)
 apply (drule_tac x=A' in bspec);apply force;
 apply (rotate_tac -1)
 apply (drule_tac x=A in bspec);apply force;
 apply (case_tac "A=A'");apply simp+;
 apply (drule context_inc_Fut);
 apply (erule disjE,erule exE);
 apply (force simp: diff_futactrefs)
 apply (rotate_tac -1, drule_tac x="Upd (ActRef B) li b" in spec);
 apply force;
 apply simp+;
 apply (subgoal_tac "fk\<notin>dom (c .\<^sub>R A)")
 apply (force simp: RequestQueue_def)
 apply (force simp: RequestQueue_def)

 apply (drule context_inc_Act);
 apply (erule disjE,erule exE);
 apply (subgoal_tac "G\<noteq>Ba");
 apply (drule diff_actrefs);
 apply force;
 apply (subgoal_tac "Ba\<in>dom c");
 apply force;
 apply (rule wf_knowsactrec);
 apply simp;
 apply (drule_tac x="Upd (ActRef B) li b" in spec);
 apply (erule exE);
 apply simp+;
 apply (rotate_tac -3)
 apply (drule_tac x=A' in bspec)
 apply force;
 apply (subgoal_tac "A\<noteq>G");
 apply simp+;
 apply force;
 apply force;
 apply (subgoal_tac "A'\<noteq>G");
 apply (case_tac "A=A'");apply simp+;

 apply (drule context_inc_Act);
 apply (erule disjE,erule exE);
 apply (subgoal_tac "G\<noteq>Ba")
 apply (rotate_tac -1)
 apply (drule diff_actrefs);
 apply force;
 apply (subgoal_tac "Ba\<in>dom c");
 apply force;
 apply (rule wf_knowsactrec);
 apply simp;
 apply (drule_tac x="Upd (ActRef B) li b" in spec);
 apply (erule exE);
 apply simp+;
 apply (rotate_tac -6)
 apply (drule_tac x=A in bspec)
 apply force;
 apply (subgoal_tac "A'\<noteq>G");
 apply simp+;
 apply force;
 apply force;
 apply simp+;
 apply (simp add: disjunct_futures_def)
 apply (rotate_tac 4)
 apply (drule_tac x=A' in bspec);apply force;
 apply (rotate_tac -1)
 apply (drule_tac x=A in bspec);apply force;
 apply (subgoal_tac "fk\<notin>dom (c .\<^sub>R A')")
 apply (force simp: RequestQueue_def)
 apply (force simp: RequestQueue_def)
 apply force;
 apply (force simp: RequestQueue_def ActiveObject_def)

 (* no cycle with G 4*)
 apply (simp add: knowsrecI)
 apply (erule exE);
 apply (case_tac "rL=[]")
 apply simp+;apply (erule knowsrec.cases)
 apply simp+;
 apply (rotate_tac -2,drule sym);
 apply simp+;
 apply (subgoal_tac "(\<exists> (E::general_context). a=E\<up>ActRef G)\<or>(\<exists> (E::general_context). b=E\<up>ActRef G)");
 apply (erule disjE);
 prefer 2
   (* case 1: atref G in A' *)
 apply (erule exE);apply simp+;
 apply (simp add: wf_conf_def wf_1b_def);apply (erule conjE)+;
 apply (rotate_tac -1, drule_tac x=A' in bspec);apply force;
 apply (rotate_tac -1, drule_tac x=fk in bspec);apply (force simp: RequestQueue_def);
 apply (rotate_tac -1, drule_tac x="E\<circ> (cUpdR (ActRef B) li (Ea))" in spec);
 apply (rotate_tac -1, drule_tac x=G in spec);
 apply (force simp: OneRequest_def);
   (* case 2: atref G in B *)
 apply (erule exE);apply simp+;
 apply (simp add: wf_conf_def wf_1a_def);apply (erule conjE)+;
 apply (rotate_tac -1, drule_tac x=B in bspec);apply force;
 apply (force simp: ActiveObject_def);

 (* 6 *)
 apply (simp add: knows_def RequestQueue_def OneRequest_def ActiveObject_def);
 apply (rotate_tac 11);
 apply (drule_tac Q= "c G = None" in conjI,assumption,drule conjunct2);
 apply (erule exE);
 apply (rotate_tac -1,drule sym,frule context_update)
 apply force;
 apply force;
 apply (drule rev_list_notNil)
 apply (erule exE)+;
 apply simp+;
 apply (drule knowsreccases_rev)
 apply (erule conjE);
 apply (subgoal_tac "x=RefFut fk") (* only fk knows G*)
 apply simp+;
 apply (case_tac x);prefer 2;
 apply simp+;
 apply (simp add: wf_conf_def wf_1b_def);apply (erule conjE)+;
 apply (rotate_tac -3,erule contrapos_pp)
 apply (simp add: knows_def RequestQueue_def OneRequest_def)
 apply (erule disjE);
 apply (rule_tac x=A' in bexI);
 apply (erule conjE,erule exE);
 apply (rule_tac x=FutureRef in bexI);
 apply (rule_tac x=Ea in exI);
 apply (rule_tac x=G in exI);
 apply (subgoal_tac "A'\<noteq>G");
 apply simp+;
 apply force;
 apply force;
 apply (subgoal_tac "A'\<noteq>G");
 apply simp+;
 apply force;
 apply force;
 apply (erule bexE);
 apply (erule conjE);
 apply (subgoal_tac "A'\<noteq>G");
 apply (subgoal_tac "A\<noteq>G");
 apply (rule_tac x=A in bexI);
 apply (erule conjE,erule exE);
 apply (rule_tac x=FutureRef in bexI);
 apply (rule_tac x=Ea in exI);
 apply (rule_tac x=G in exI);
 apply simp+;apply auto;
 apply (case_tac "A=A'");apply force;apply simp+;
 apply (case_tac "A=A'");  apply force; apply force;

 apply (case_tac "ActivityRef=G");
 apply simp+;
 apply (simp add: knows_def RequestQueue_def OneRequest_def ActiveObject_def)
 apply auto;
 apply (subgoal_tac "(\<exists> (E::general_context). a=E\<up>ActRef G)\<or>(\<exists> (E::general_context). b=E\<up>ActRef G)");
 prefer 2;
 apply (rotate_tac -1,drule sym,frule context_update)
 apply force;
 apply (erule disjE,erule exE)
 (* a cannot reference G *)
 apply (simp add: wf_conf_def wf_1a_def);apply (erule conjE)+;
 apply (rotate_tac -4,erule contrapos_pp)
 apply simp+;
 apply (rule_tac x=B in bexI);
 apply (rule_tac x=Eb in exI);
 apply (rule_tac x=G in exI);
 apply (simp add: ActiveObject_def)
 apply force;
 apply (force simp: RequestQueue_def)
 (* b cannot reference G *)
 apply (erule exE);
  apply (simp add: wf_conf_def wf_1b_def);apply (erule conjE)+;
 apply (rotate_tac -3,erule contrapos_pp)
 apply simp+;
 apply (rule_tac x=A' in bexI);
 apply (rule_tac x=fk in bexI);
 apply (rule_tac x="(E \<circ> (cUpdR (ActRef B) li Eb))" in exI);
 apply (rule_tac x=G in exI);
 apply (simp add: OneRequest_def)
 apply force;
 apply (force simp: RequestQueue_def)
 apply force;
 apply (rotate_tac -2);
 apply (simp add: wf_conf_def wf_1a_def knows_def RequestQueue_def OneRequest_def ActiveObject_def);apply (erule conjE)+;
 apply (rotate_tac -6,erule contrapos_pp)
 apply simp+; 
 apply (rule_tac x=ActivityRef in bexI); apply (erule exE);
 apply (rule_tac x=Ea in exI);
 apply (rule_tac x=G in exI);
 apply (case_tac "ActivityRef=A'");
 apply force;apply force;
 apply force;

(* REMOTE METHOD INVOCVATION  *)

 (* preliminary: *)
 apply (subgoal_tac "\<forall> A \<in> dom c. A\<noteq> A' \<longrightarrow> (fk\<notin> dom (c .\<^sub>R A) )");
 prefer 2;apply (simp add: disjunct_futures_def); apply (force simp: RequestQueue_def);


 apply (subgoal_tac "\<forall> r. ((RefFut fk,r) \<in> knows (c(A' \<mapsto> ((c .\<^sub>R A')(fk \<mapsto> E \<up> FutRef fn), c .\<^sub>A A'))) \<longrightarrow> (((RefFut fk,r) \<in> knows c)\<or>(r=RefFut fn)))");
 prefer 2; apply (rule allI,rule impI);
 apply (simp add: knows_def RequestQueue_def OneRequest_def);
 apply (case_tac r)
 apply auto;
 apply (drule context_inc_Act)
 apply (erule disjE,erule exE);
 apply (force simp: diff_actfutrefs);
 apply (drule_tac x=A' in bspec); apply force;
 apply (drule_tac x="Call (ActRef B) li p" in spec);
 apply (erule exE); apply simp+;
 apply (rule_tac x=A' in bexI);apply force;apply force;
 
 apply (subgoal_tac "A= A'")
 apply simp;
  (* again :-( *)
 apply (drule context_inc_Act)
 apply (erule disjE,erule exE);
 apply (force simp: diff_actfutrefs);
 apply (drule_tac x=A' in bspec); apply force;
 apply (drule_tac x="Call (ActRef B) li p" in spec);
 apply (erule exE); apply simp+;
 apply (rule_tac x=A' in bexI);apply force;apply force;
 apply (rotate_tac 5,drule_tac x=A in bspec);
 apply force;apply force;


 apply (frule context_inc_Fut)
 apply (erule disjE,erule exE); 
 apply (rotate_tac 10)
 apply (drule not_sym)
 apply (drule diff_futrefs)
 apply (force simp: diff_futrefs);
 apply (drule_tac x=A' in bspec); apply force;
 apply (drule_tac x="Call (ActRef B) li p" in spec);
 apply (erule exE); apply simp+;
 apply (rule_tac x=A' in bexI);apply force;apply force;
 
 apply (subgoal_tac "A= A'")
 apply simp;
  (* again :-( *)
 apply (drule context_inc_Fut)
 apply (erule disjE,erule exE); 
 apply (rotate_tac 10)
 apply (drule not_sym)
 apply (drule diff_futrefs)
 apply (force simp: diff_futrefs);
 apply (drule_tac x=A' in bspec); apply force;
 apply (drule_tac x="Call (ActRef B) li p" in spec);
 apply (erule exE); apply simp+;
 apply (rule_tac x=A' in bexI);apply force;apply force;
 apply (rotate_tac 5,drule_tac x=A in bspec);
 apply force;apply force;


 (* a crucial subgoal *)
 apply (subgoal_tac " \<forall> r .(r,RefFut fn )\<in> knows (c(A' \<mapsto> (m'(fk \<mapsto> E \<up> FutRef fn), a'), B \<mapsto> (mb(fn \<mapsto> Call a li p), a)))\<longrightarrow> r= RefFut fk" );
 (*rewriting*)  apply (subgoal_tac "(c(A' \<mapsto> (m'(fk \<mapsto> E \<up> FutRef fn), a'), B \<mapsto> (mb(fn \<mapsto> Call a li p), a))) = (c(A' \<mapsto> ((c .\<^sub>R A')(fk \<mapsto>  E \<up> FutRef fn), c .\<^sub>A A')))(B \<mapsto> ((c .\<^sub>R B)(fn \<mapsto> Call a li p), c .\<^sub>A B))" );
                   prefer 2;apply (force simp: RequestQueue_def ActiveObject_def OneRequest_def);
		apply (subgoal_tac "(c .\<^sub>R B) =((c(A' \<mapsto> ((c .\<^sub>R A')(fk \<mapsto> E \<up> FutRef fn), c .\<^sub>A A'))) .\<^sub>R B) " );
                   prefer 2;apply (simp(no_asm_simp) add: RequestQueue_def ActiveObject_def);
		apply (subgoal_tac "(c .\<^sub>A B) =((c(A' \<mapsto> ((c .\<^sub>R A')(fk \<mapsto> E \<up> FutRef fn), c .\<^sub>A A'))) .\<^sub>A B) " );
                prefer 2;apply (simp(no_asm_simp) add: RequestQueue_def ActiveObject_def);
 apply (simp);
 apply (rule_tac c="(c(A' \<mapsto> ((c .\<^sub>R A')(fk \<mapsto> E \<up> FutRef fn), c .\<^sub>A A')))" and  A=B in nocycle_fut_upd)
   (* 6 *)
 apply (rule nocycle_fut_upd)
 apply simp+;apply force;
 apply (rotate_tac -2);
 apply (drule sym,drule sym, simp+);
   (*eliminate the rewriting subgoals *) 
                 apply (thin_tac " (c(A' \<mapsto> ((c .\<^sub>R A')(fk \<mapsto> E \<up> FutRef fn), c .\<^sub>A A')) .\<^sub>A B)=(c .\<^sub>A B) ");
		 apply (thin_tac "(c(A' \<mapsto> ((c .\<^sub>R A')(fk \<mapsto> E \<up> FutRef fn), c .\<^sub>A A')) .\<^sub>R B) = (c .\<^sub>R B)");
		 apply (thin_tac "c(A' \<mapsto> (m'(fk \<mapsto> E \<up> FutRef fn), a'), B \<mapsto> (mb(fn \<mapsto> Call a li p), a)) = c(A' \<mapsto> ((c .\<^sub>R A')(fk \<mapsto> E \<up> FutRef fn), c .\<^sub>A A'), B \<mapsto>
        ((c .\<^sub>R B)(fn \<mapsto> Call a li p), c .\<^sub>A B))");
  (* 6 *)
 apply (simp add: knowsrecI) 
 apply (erule contrapos_pp)
 apply simp+;
 apply (erule exE);
 apply (drule minimal_path);
 apply (erule exE,(erule conjE)+, simp+);
 apply (simp add: nocycle_def knowsrecI);
 apply (erule knowsrec.cases  );
 apply simp+;
 apply (rotate_tac -3,drule sym,drule sym)
 apply simp+;
 apply (rule_tac x=r in exI)
 apply (rule_tac x="[]" in exI)
 apply (intro knowsrec.intros);
 apply simp+;
 apply (drule_tac x="RefFut fk" in spec);
 apply simp+;
 apply (erule disjE);
 apply force;
 apply (drule_tac x=A' in bspec);apply force;apply force;

 apply (rule_tac x=r in exI)
 apply (simp,(erule conjE)+);
 apply (rule_tac x="r'#rL" in exI)
 apply (intro knowsrec.intros);
 apply (drule_tac x="r'" in spec);
 apply simp+;
 apply (drule pre_nocycle_fut_upd');apply force;apply force;
 apply (erule disjE); apply simp+;
 apply (drule wf_knowsfutrec);
 apply (erule bexE);
 apply (drule_tac x= A in bspec);
 apply simp+;
 apply (simp add: RequestQueue_def);

  (* 5 *)
 apply (force);
 apply (rotate_tac -2,drule sym,drule sym, simp+);
   (*eliminate the rewriting subgoals *) 
                 apply (thin_tac " (c(A' \<mapsto> ((c .\<^sub>R A')(fk \<mapsto> E \<up> FutRef fn), c .\<^sub>A A')) .\<^sub>A B)=(c .\<^sub>A B) ");
		 apply (thin_tac "(c(A' \<mapsto> ((c .\<^sub>R A')(fk \<mapsto> E \<up> FutRef fn), c .\<^sub>A A')) .\<^sub>R B) = (c .\<^sub>R B)");

 (* 4 *)
 apply (thin_tac "c(A' \<mapsto> (m'(fk \<mapsto> E \<up> FutRef fn), a'), B \<mapsto> (mb(fn \<mapsto> Call a li p), a)) = c(A' \<mapsto> ((c .\<^sub>R A')(fk \<mapsto> E \<up> FutRef fn), c .\<^sub>A A'), B \<mapsto>
        ((c .\<^sub>R B)(fn \<mapsto> Call a li p), c .\<^sub>A B))")
 apply (subgoal_tac 
" \<forall> r. ((RefFut fn,r)\<in> knows (c(A' \<mapsto> ((c .\<^sub>R A')(fk \<mapsto> E \<up> FutRef fn), c .\<^sub>A A'), B \<mapsto> ((c .\<^sub>R B)(fn \<mapsto> Call a li p), c .\<^sub>A B))) 
                \<longrightarrow>((RefFut fk,r)\<in> knows c)\<or>(RefAct B,r)\<in> knows c)");
 apply (erule contrapos_pp);
 apply (simp add: knowsrecI nocycle_def)+;
 apply (erule exE,case_tac "rL=[]");
 apply (erule knowsrec.cases)
 apply (rotate_tac -4,drule sym,drule sym, simp+)
 apply (rotate_tac -6)
 apply (drule_tac x="RefFut fn" in spec)
 apply simp+;
 apply (drule_tac x=A' in bspec);apply force;apply force;
 apply simp+;
  (* 5 *)
 apply (drule rev_list_notNil)
 apply (erule exE)+;
 apply simp+;
 apply (drule knowsreccases_rev)
 apply (erule conjE)
 apply (subgoal_tac "x=RefFut fk");prefer 2; apply force;
 apply (rotate_tac -1);
 apply (thin_tac "\<forall>r. (r, RefFut fn)
            \<in> knows
               (c(A' \<mapsto> ((c .\<^sub>R A')(fk \<mapsto> E \<up> FutRef fn), c .\<^sub>A A'), B \<mapsto>
                ((c .\<^sub>R B)(fn \<mapsto> Call a li p), c .\<^sub>A B))) \<longrightarrow>
            r = RefFut fk");
 apply (simp);apply (thin_tac "x=RefFut fk")
 apply (rotate_tac -1,drule minimal_path,erule exE)
 apply (erule conjE);
 apply (erule knowsrec.cases)
 apply (rotate_tac -4,drule sym,drule sym, simp+)
 apply (rotate_tac 13)
 apply (drule_tac x="RefFut fk" in spec)
 apply simp+;
 apply (erule disjE);
 apply (rule_tac x="RefFut fk" in exI);
 apply (rule_tac x="[]" in exI);
 apply (force intro: knowsrec.intros);
 apply (rule_tac x="RefFut fk" in exI);
 apply (rule_tac x="[RefAct B]" in exI);
 apply (intro  knowsrec.intros);
 apply (rule conjI);
 apply (thin_tac "\<forall>A\<in>dom c. A \<noteq> A' \<longrightarrow> fk \<notin> dom (c .\<^sub>R A)")
 apply (simp(no_asm_simp) add: knows_def OneRequest_def RequestQueue_def);
 apply (rule_tac x=A' in bexI);apply (rule conjI, force);
 apply (rule_tac x="E\<circ>(cCallL cHole li p)" in exI)
 apply auto;
  apply (force intro: knowsrec.intros);
 apply (rotate_tac -4);
 apply (subgoal_tac "(r', RefFut fk, rLb) \<in> knowsrec ((c(A' \<mapsto> ((c .\<^sub>R A')(fk \<mapsto> E \<up> FutRef fn), c .\<^sub>A A'))) (B \<mapsto> (((c(A' \<mapsto> ((c .\<^sub>R A')(fk \<mapsto> E \<up> FutRef fn), c .\<^sub>A A') )) .\<^sub>R B)(fn \<mapsto> Call a li p), (c(A' \<mapsto> ((c .\<^sub>R A')(fk \<mapsto> E \<up> FutRef fn), c .\<^sub>A A'))) .\<^sub>A B)))");
 prefer 2; apply (force simp: RequestQueue_def ActiveObject_def) 
 apply (drule pre_nocycle_fut_upd')
 apply force;apply force;
 apply (drule pre_nocycle_fut_upd')
 apply force;apply force;
 apply (rotate_tac 15);
 apply (drule_tac x=r' in spec)
 apply simp+;
 apply (erule disjE)
 apply (rule_tac x="RefFut fk" in exI)
 apply (rule_tac x="r'#rLb" in exI);
 apply (force intro: knowsrec.intros);
 apply (rule_tac x="RefFut fk" in exI)
 apply (rule_tac x="RefAct B #r'#rLb" in exI);
 apply (subgoal_tac "(RefFut fk,RefAct B)\<in> knows c")
 apply (force intro: knowsrec.intros);
 apply (simp(no_asm_simp) add: knows_def OneRequest_def RequestQueue_def);
 apply (rule_tac x=A' in bexI);apply (rule conjI, force);
 apply (rule_tac x="E\<circ>(cCallL cHole li p)" in exI)
 apply simp+;apply force;
(* how can fn ref a fut or act? 4 *)
 apply (case_tac r);
 apply (subgoal_tac "\<exists> E::general_context. Call a li p= E\<up>ActRef ActivityRef");
 apply (subgoal_tac "(\<exists> E::general_context. a = E\<up>ActRef ActivityRef)\<or>(\<exists> E::general_context. p= E\<up>ActRef ActivityRef)");
 apply (erule disjE,(erule exE)+);
 apply simp+;
 apply (thin_tac "\<forall>A\<in>dom c. A \<noteq> A' \<longrightarrow> fk \<notin> dom (c .\<^sub>R A)");
 apply (thin_tac "\<forall>r. (r, RefFut fn)
            \<in> knows
               (c(A' \<mapsto> ((c .\<^sub>R A')(fk \<mapsto> E \<up> FutRef fn), c .\<^sub>A A'), B \<mapsto>
                ((c .\<^sub>R B)(fn \<mapsto> Ea \<up> ActRef ActivityRef), c .\<^sub>A B))) \<longrightarrow>  r = RefFut fk");
 apply (thin_tac "(RefFut fn, RefAct ActivityRef) \<in> knows
           (c(A' \<mapsto> ((c .\<^sub>R A')(fk \<mapsto> E \<up> FutRef fn), c .\<^sub>A A'), B \<mapsto>
            ((c .\<^sub>R B)(fn \<mapsto> Ea \<up> ActRef ActivityRef), c .\<^sub>A B)))");
 apply (force simp: knows_def RequestQueue_def OneRequest_def ActiveObject_def);
 apply ((erule exE)+);
 apply (simp add: knows_def);
 apply (simp add: RequestQueue_def OneRequest_def);
 apply (rule_tac x=A' in bexI)
 apply (rule conjI, force);
 apply (rule_tac x="E\<circ>(cCallR (ActRef B) li Eb)" in exI)
 apply simp;apply force; 
 apply (erule exE);
 apply (rotate_tac -1,drule sym)
 apply (frule context_call_1);
 apply (erule disjE)+;
 apply simp+;

 apply (simp add: knows_def); 
 apply (simp add: OneRequest_def RequestQueue_def)
 apply (subgoal_tac "fn\<noteq>fk");apply simp+;
 apply (erule disjE)+
 apply force;
 apply (erule disjE)+
 apply (drule_tac x=A' in bspec);
 apply force;apply force
 apply (erule bexE)
 apply (drule_tac x=A in bspec);
 apply force;
 apply (case_tac "A=B");apply force;apply simp+;
 apply (case_tac "A=A'");apply force;apply simp+;
 apply (drule_tac x=A' in bspec);apply force;
 apply (rotate_tac 8);
 apply (drule_tac Q= "wf_conf c" in conjI,assumption,drule conjunct2);
 apply (drule_tac Q= "wf_conf c" in conjI,assumption,drule conjunct2);
 apply force;

 (* case futref 4 *)
 apply (subgoal_tac "\<exists> E::general_context. Call a li p= E\<up>FutRef FutureRef");
 apply (subgoal_tac "(\<exists> E::general_context. a = E\<up>FutRef FutureRef)\<or>(\<exists> E::general_context. p= E\<up>FutRef FutureRef)");
 apply (erule disjE,(erule exE)+);
 apply ( simp add: knows_def);
 apply (simp add: RequestQueue_def OneRequest_def ActiveObject_def);
  apply (rotate_tac 10);
 apply (drule_tac Q= "wf_conf c" in conjI,assumption,drule conjunct2);
 apply (drule_tac Q= "wf_conf c" in conjI,assumption,drule conjunct2);
 apply force;
 apply ((erule exE)+);
 apply (simp add: knows_def);apply (simp add: RequestQueue_def OneRequest_def);
 apply (rule_tac x=A' in bexI)
 apply (rule conjI, force);
 apply (rule_tac x="E\<circ>(cCallR (ActRef B) li Eb)" in exI)
 apply simp;apply force; 
 apply (erule exE);
 apply (rotate_tac -1,drule sym)
 apply (frule context_call_1);
 apply (erule disjE)+;
 apply simp+;

 apply (simp add: knows_def);apply (simp add: OneRequest_def RequestQueue_def)
 apply (subgoal_tac "fn\<noteq>fk");apply simp+;
 apply (erule disjE)+
 apply force;
 apply (erule disjE)+
 apply (drule_tac x=A' in bspec);
 apply force;apply force
 apply (erule bexE)
 apply (drule_tac x=A in bspec);
 apply force;
 apply (case_tac "A=B");apply force;apply simp+;
 apply (case_tac "A=A'");apply force;apply simp+;
 apply (drule_tac x=A' in bspec);apply force; apply (rotate_tac 8);
 apply (drule_tac Q= "wf_conf c" in conjI,assumption,drule conjunct2);
 apply (drule_tac Q= "wf_conf c" in conjI,assumption,drule conjunct2);
 apply force;

 apply (case_tac r);
 apply (simp add: knows_def OneRequest_def RequestQueue_def ActiveObject_def wf_conf_def wf_2a_def)
 apply (erule conjE)+;apply (rotate_tac -2);apply (erule contrapos_pp);
 apply simp+;
 apply (erule exE)
 apply (rule_tac x=ActivityRef in bexI);
 apply (rule_tac x=Ea in exI);
 apply (rule_tac x=fn in exI); 
 apply (case_tac "ActivityRef=B");apply force;
 apply (case_tac "ActivityRef=A'");apply force;apply force;
 apply force;
 apply (simp add: knows_def OneRequest_def RequestQueue_def ActiveObject_def wf_conf_def wf_2b_def)
 apply (erule conjE)+; apply (subgoal_tac "FutureRef\<noteq> fn")

 apply (rotate_tac -2);apply (erule contrapos_pp);
 apply simp+;
 apply (erule disjE)+;
 apply (erule conjE, erule exE);
 apply (rule_tac x=B in bexI);
 apply (rule_tac x=FutureRef in bexI); 
 apply (rule_tac x=Ea in exI);
 apply (rule_tac x=fn in exI);
 apply (case_tac "ActivityRef=B");apply force;
 apply (case_tac "ActivityRef=A'");apply force;apply force;
 apply force;
 apply force;
 apply (erule disjE)+;
 apply (erule conjE, (erule exE)+);
 apply (rule_tac x=A' in bexI);
 apply (rule_tac x=FutureRef in bexI); 
 apply (rule_tac x=Ea in exI);
 apply (rule_tac x=fn in exI);
 apply force;
 apply force;
 apply force;
 
 apply ((erule bexE),erule conjE,(erule exE)+);
 apply (case_tac "A=B");
 apply simp+;
 apply (rule_tac x=A in bexI);
 apply (rule_tac x=FutureRef in bexI); 
 apply (rule_tac x=Ea in exI);
 apply (rule_tac x=fn in exI);
 apply force;
 apply force;
 apply force;

 apply simp+;
 apply (case_tac "A=A'");
 apply simp+;
 apply (rule_tac x=A in bexI);
 apply (rule_tac x=FutureRef in bexI); 
 apply (rule_tac x=Ea in exI);
 apply (rule_tac x=fn in exI);
 apply force;
 apply force;
 apply force;
 apply simp+;
 apply (rule_tac x=A in bexI);
 apply (rule_tac x=FutureRef in bexI); 
 apply (rule_tac x=Ea in exI);
 apply (rule_tac x=fn in exI);
 apply force;
 apply force;
 apply force;

(* how can fn ref a fut? 3 copy-paste *)

 (* case futref 3*)
 apply clarify;
 apply simp+;
 apply (rotate_tac 9);
 apply (drule_tac Q= "nocycle c" in conjI,assumption,drule conjunct2);
 apply (subgoal_tac " (\<exists> E::general_context. a =E\<up>(FutRef fn))\<or>(\<exists> E::general_context. p =E\<up>(FutRef fn))" )
 apply (drule_tac Q= "nocycle c" in conjI,assumption,drule conjunct2);
 apply (simp add: knowsrecI nocycle_def)+;
 apply (erule disjE,(erule exE)+);
 apply (simp add: wf_2a_def);
 apply (rotate_tac 2,erule contrapos_pp)
 apply simp+;
 apply (rule_tac x="B" in bexI)
 apply (rule_tac x=Ea in exI)
 apply (rule_tac x=fn in exI)
 apply (force simp: RequestQueue_def ActiveObject_def);apply force;
 apply (erule exE);
 apply (rotate_tac 3,erule contrapos_pp)
 apply simp+;
 apply (rule_tac x=A' in bexI)
 apply (rule_tac x=fk in bexI)
 apply (rule_tac x="E\<circ>(cCallR (ActRef B) li Ea)" in exI)
 apply (rule_tac x=fn in exI)
 apply (force simp: RequestQueue_def ActiveObject_def);apply force;apply force;

 apply (subgoal_tac " \<exists>E::general_context. Call a li p = E\<up>(FutRef fn)");
 apply (erule exE)
 apply (rotate_tac -1,drule sym)
 apply (frule context_call_1);
 apply (erule disjE)+;
 apply simp+;
 apply force;
 apply (erule disjE)+;
 apply simp+;
 apply (erule disjE)+;
 apply (rotate_tac 4);
 apply (drule_tac x=A' in bspec);apply force;apply force; 
 apply (erule bexE);
 apply (case_tac "A=B");apply simp;
 apply (subgoal_tac "fn\<noteq>fk");apply simp+;
 apply (case_tac "A=A'"); apply (rotate_tac 4);
 apply (drule_tac x=A' in bspec);
 apply force;
 apply simp+;
 apply (rotate_tac 4);
 apply (drule_tac x=A' in bspec);
 apply force;
 apply (erule conjE);
 apply force;

(* SELF REQUEST  2 *)

 (* preliminary: *)
 apply (subgoal_tac "\<forall> A \<in> dom c. A\<noteq> A' \<longrightarrow> (fk\<notin> dom (c .\<^sub>R A) )");
 prefer 2;apply (simp add: disjunct_futures_def); apply (force simp: RequestQueue_def);


 apply (subgoal_tac "\<forall> r. ((RefFut fk,r) \<in> knows (c(A' \<mapsto> ((c .\<^sub>R A')(fk \<mapsto> E \<up> FutRef fn), c .\<^sub>A A'))) \<longrightarrow> (((RefFut fk,r) \<in> knows c)\<or>(r=RefFut fn)))");
 prefer 2; apply (rule allI,rule impI);
 apply (simp add: knows_def);apply (simp add: RequestQueue_def OneRequest_def);
 apply (case_tac r)
 apply auto;
 apply (drule context_inc_Act)
 apply (erule disjE,erule exE);
 apply (force simp: diff_actfutrefs);
 apply (drule_tac x=A' in bspec); apply force;
 apply (drule_tac x="Call (ActRef A') li p" in spec);
 apply (erule exE); apply simp+;
 apply (rule_tac x=A' in bexI);apply force;apply force;
 
 apply (subgoal_tac "A= A'")
 apply simp;
  (* again :-( *)
 apply (drule context_inc_Act)
 apply (erule disjE,erule exE);
 apply (force simp: diff_actfutrefs);
 apply (drule_tac x=A' in bspec); apply force;
 apply (drule_tac x="Call (ActRef A') li p" in spec);
 apply (erule exE); apply simp+;
 apply (rule_tac x=A' in bexI);apply force;apply force;
 apply (rotate_tac 5,drule_tac x=A in bspec);
 apply force;apply force;


 apply (frule context_inc_Fut)
 apply (erule disjE,erule exE); 
 apply (rotate_tac 10)
 apply (drule not_sym)
 apply (drule diff_futrefs)
 apply (force simp: diff_futrefs);
 apply (drule_tac x=A' in bspec); apply force;
 apply (drule_tac x="Call (ActRef A') li p" in spec);
 apply (erule exE); apply simp+;
 apply (rule_tac x=A' in bexI);apply force;apply force;
 
 apply (subgoal_tac "A= A'")
 apply simp;
  (* again :-( *)
 apply (drule context_inc_Fut)
 apply (erule disjE,erule exE); 
 apply (rotate_tac 10)
 apply (drule not_sym)
 apply (drule diff_futrefs)
 apply (force simp: diff_futrefs);
 apply (drule_tac x=A' in bspec); apply force;
 apply (drule_tac x="Call (ActRef A') li p" in spec);
 apply (erule exE); apply simp+;
 apply (rule_tac x=A' in bexI);apply force;apply force;
 apply (rotate_tac 5,drule_tac x=A in bspec);
 apply force;apply force;


 (* a crucial subgoal *)
 apply (subgoal_tac " \<forall> r .(r,RefFut fn )\<in> knows (c(A' \<mapsto> (m'(fk \<mapsto> E \<up> FutRef fn,fn \<mapsto> Call a li p), a)))\<longrightarrow> r= RefFut fk" );
 (*rewriting*)  apply (subgoal_tac "(c(A' \<mapsto> (m'(fk \<mapsto> E \<up> FutRef fn, fn \<mapsto> Call a li p), a))) = (c(A' \<mapsto> ((c .\<^sub>R A')(fk \<mapsto>  E \<up> FutRef fn), c .\<^sub>A A')))(A' \<mapsto> (((c(A' \<mapsto> ((c .\<^sub>R A')(fk \<mapsto>  E \<up> FutRef fn), c .\<^sub>A A'))) .\<^sub>R A')(fn \<mapsto> Call a li p), (c(A' \<mapsto> ((c .\<^sub>R A')(fk \<mapsto>  E \<up> FutRef fn), c .\<^sub>A A'))) .\<^sub>A A'))" ); 
                   prefer 2;apply (force simp: RequestQueue_def ActiveObject_def OneRequest_def);
 apply (rotate_tac -1, erule ssubst);
 apply (rule_tac c="(c(A' \<mapsto> ((c .\<^sub>R A')(fk \<mapsto>  E \<up> FutRef fn), c .\<^sub>A A')))" and  A=A' in nocycle_fut_upd)
   (* 5 *)
 apply (rule nocycle_fut_upd)
 apply simp+;apply force;
  (* 5 *)
 apply (simp add: knowsrecI) 
 apply (erule contrapos_pp)
 apply simp+;
 apply (erule exE);
 apply (drule minimal_path);
 apply (erule exE,(erule conjE)+, simp+);
 apply (simp add: nocycle_def knowsrecI);
 apply (erule knowsrec.cases  );
 apply simp+;
 apply (rotate_tac -3,drule sym)
 apply simp+;
 apply (rule_tac x=r in exI)
 apply (rule_tac x="[]" in exI)
 apply (intro knowsrec.intros);
 apply simp+;
 apply (drule_tac x="RefFut fk" in spec);
 apply simp+;
 apply (erule disjE);
 apply force;
 apply (drule_tac x=A' in bspec);apply force;apply force;

 apply (rule_tac x=r in exI)
 apply (simp,(erule conjE)+);
 apply (rule_tac x="r'#rL" in exI)
 apply (intro knowsrec.intros);
 apply (drule_tac x="r'" in spec);
 apply simp+;
 apply (drule pre_nocycle_fut_upd');apply force;apply force;
 apply (erule disjE); apply simp+;
 apply (drule wf_knowsfutrec);
 apply (erule bexE);
 apply (drule_tac x= A in bspec);
 apply simp+;
 apply (simp add: RequestQueue_def);

  (* 4 *)
 apply (force);

 (* 3 *)
 apply (simp add: RequestQueue_def ActiveObject_def)
 apply (subgoal_tac 
" \<forall> r. ((RefFut fn,r)\<in> knows (c(A' \<mapsto> (m'(fk \<mapsto> E \<up> FutRef fn, fn \<mapsto> Call a li p), a))) 
                \<longrightarrow>((RefFut fk,r)\<in> knows c)\<or>(RefAct A',r)\<in> knows c)");
 apply (erule contrapos_pp);
 apply (simp add: knowsrecI nocycle_def)+;
 apply (erule exE,case_tac "rL=[]");
 apply (erule knowsrec.cases)
 apply (rotate_tac -4,drule sym,drule  sym, simp+)
 apply (rotate_tac -5)
 apply (drule_tac x="RefFut fn" in spec)
 apply simp+;
 apply (drule_tac x=A' in bspec);apply force;apply force;
 apply simp+;
  (* 4 *)
 apply (drule rev_list_notNil)
 apply (erule exE)+;
 apply simp+;
 apply (drule knowsreccases_rev)
 apply (erule conjE)
 apply (subgoal_tac "x=RefFut fk");prefer 2; apply force;
 apply (rotate_tac -1,simp);
 apply (rotate_tac -1,drule minimal_path,erule exE)
 apply (erule conjE);
 apply (erule knowsrec.cases)
 apply (rotate_tac -4,drule  sym, drule sym, simp+)
 apply (rotate_tac -4)
 apply (drule_tac x="RefFut fk" in spec)
 apply simp+;
 apply (erule disjE);
 apply (rule_tac x="RefFut fk" in exI);
 apply (rule_tac x="[]" in exI);
 apply (force intro: knowsrec.intros);
 apply (rule_tac x="RefFut fk" in exI);
 apply (rule_tac x="[RefAct A']" in exI);
 apply (intro  knowsrec.intros);
 apply (rule conjI);
 apply (simp(no_asm_simp) add: knows_def); 
  apply (simp(no_asm_simp) add: OneRequest_def RequestQueue_def);

 apply (rule_tac x=A' in bexI);apply (rule conjI, force);
 apply (rule_tac x="E\<circ>(cCallL cHole li p)" in exI)
 apply auto;
  apply (force intro: knowsrec.intros);
 apply (subgoal_tac "(r', RefFut fk, rLb) \<in> knowsrec ((c(A' \<mapsto> ((c .\<^sub>R A')(fk \<mapsto>  E \<up> FutRef fn), c .\<^sub>A A')))(A' \<mapsto> (((c(A' \<mapsto> ((c .\<^sub>R A')(fk \<mapsto>  E \<up> FutRef fn), c .\<^sub>A A'))) .\<^sub>R A')(fn \<mapsto> Call a li p), (c(A' \<mapsto> ((c .\<^sub>R A')(fk \<mapsto>  E \<up> FutRef fn), c .\<^sub>A A'))) .\<^sub>A A')))");
 prefer 2; apply (force simp: RequestQueue_def ActiveObject_def) 
 apply (drule pre_nocycle_fut_upd')
 apply force;apply force;
 apply (drule pre_nocycle_fut_upd')
 apply force;apply force;
 apply (rotate_tac -8);
 apply (drule_tac x=r' in spec)
 apply simp+;
 apply (erule disjE)
 apply (rule_tac x="RefFut fk" in exI)
 apply (rule_tac x="r'#rLb" in exI);
 apply (force intro: knowsrec.intros);
 apply (rule_tac x="RefFut fk" in exI)
 apply (rule_tac x="RefAct A' #r'#rLb" in exI);
 apply (subgoal_tac "(RefFut fk,RefAct A')\<in> knows c")
 apply (force intro: knowsrec.intros);
 apply (simp(no_asm_simp) add: knows_def );
 apply (simp(no_asm_simp) add: OneRequest_def RequestQueue_def);
 apply (rule_tac x=A' in bexI);apply (rule conjI, force);
 apply (rule_tac x="E\<circ>(cCallL cHole li p)" in exI)
 apply simp+;apply force;
(* how can fn ref a fut or act? 3 *)
 apply (case_tac r);
 apply (subgoal_tac "\<exists> E::general_context. Call a li p= E\<up>ActRef ActivityRef");
 apply (subgoal_tac "(\<exists> E::general_context. a = E\<up>ActRef ActivityRef)\<or>(\<exists> E::general_context. p= E\<up>ActRef ActivityRef)");
 apply (erule disjE,(erule exE)+);
 apply (thin_tac "\<forall>A\<in>dom c. A \<noteq> A' \<longrightarrow> fk \<notin> dom (fst (the (c A)))");
 apply (simp add: knows_def );
 apply (simp add: RequestQueue_def OneRequest_def ActiveObject_def);
 apply force;
 apply ((erule exE)+);
 apply (simp add: knows_def);
 apply (simp add: RequestQueue_def OneRequest_def);
 apply (rule_tac x=A' in bexI)
 apply (rule conjI, force);
 apply (rule_tac x="E\<circ>(cCallR (ActRef A') li Eb)" in exI)
 apply simp;apply force; 
 apply (erule exE);
 apply (rotate_tac -1,drule sym)
 apply (frule context_call_1);
 apply (erule disjE)+;
 apply simp+;

 apply (simp add: knows_def);apply (simp add: OneRequest_def RequestQueue_def)
 apply (erule disjE)+
 apply force;
 apply (erule bexE)
 apply (erule conjE,erule exE);
 apply (drule_tac x=A in bspec);
 apply force;
 apply (case_tac "A=A'");prefer 2;apply force;
 apply simp+;
 apply force
(* apply (drule_tac x=A' in bspec);
 apply force;
 apply force;*)

 (* case futref 3 *)
 apply (subgoal_tac "\<exists> E::general_context. Call a li p= E\<up>FutRef FutureRef");
 apply (subgoal_tac "(\<exists> E::general_context. a = E\<up>FutRef FutureRef)\<or>(\<exists> E::general_context. p= E\<up>FutRef FutureRef)");
 apply (erule disjE,(erule exE)+);
 apply ( simp add: knows_def); apply (simp add: RequestQueue_def OneRequest_def ActiveObject_def); apply force;
 apply ((erule exE)+);
 apply (simp add: knows_def); apply (simp add: RequestQueue_def OneRequest_def);
 apply (rule_tac x=A' in bexI)
 apply (rule conjI, force);
 apply (rule_tac x="E\<circ>(cCallR (ActRef A') li Eb)" in exI)
 apply simp;apply force; 
 apply (erule exE);
 apply (rotate_tac -1,drule sym)
 apply (frule context_call_1);
 apply (erule disjE)+;
 apply simp+;

 apply (simp add: knows_def);  apply (simp add: OneRequest_def RequestQueue_def)
 apply (erule disjE)+
 apply force;
 apply (erule bexE)
 apply (erule conjE,erule exE);
 apply (drule_tac x=A in bspec);
 apply force;
 apply (case_tac "A=A'");prefer 2;apply force;
 apply simp+;
 apply force

 (* 2 *)
 apply (case_tac r);
 apply (simp add: knows_def); apply (simp add: OneRequest_def RequestQueue_def ActiveObject_def wf_conf_def wf_2a_def)
 apply (erule conjE)+;apply (rotate_tac -2);apply (erule contrapos_pp);
 apply simp+;
 apply (erule exE)
 apply (rule_tac x=ActivityRef in bexI);
 apply (rule_tac x=Ea in exI);
 apply (rule_tac x=fn in exI); 
 apply (case_tac "ActivityRef=A'");apply force;apply force;
 apply force;

(* subgoal only fk references fn 2*)
 apply (simp add: knows_def);  apply (simp add: OneRequest_def RequestQueue_def ActiveObject_def wf_conf_def wf_2b_def)
 apply (erule conjE)+; apply (subgoal_tac "FutureRef\<noteq> fn")

 apply (rotate_tac -2);apply (erule contrapos_pp);
 apply simp+;
 apply (erule disjE)+;
 apply (erule conjE, erule exE);
 apply (rule_tac x=A' in bexI);
 apply (rule_tac x=FutureRef in bexI); 
 apply (rule_tac x=Ea in exI);
 apply (rule_tac x=fn in exI);
 apply (case_tac "ActivityRef=A'");apply force;apply force;
 apply force;
 apply force;
 
 apply ((erule bexE),erule conjE,(erule exE)+);
 apply (case_tac "A=A'");
 apply simp+;
 apply (rule_tac x=A in bexI);
 apply (rule_tac x=FutureRef in bexI); 
 apply (rule_tac x=Ea in exI);
 apply (rule_tac x=fn in exI);
 apply force;
 apply force;
 apply force;

 apply simp+;
 apply (case_tac "A=A'");
 apply simp+;
 apply (rule_tac x=A in bexI);
 apply (rule_tac x=FutureRef in bexI); 
 apply (rule_tac x=Ea in exI);
 apply (rule_tac x=fn in exI);
 apply force;
 apply force;
 apply force;

(* how can fn ref a fut? 2 ~copy-paste *)

 (* case futref 2*)
 apply clarify;
 apply simp+;
 apply (rotate_tac 7);
 apply (drule_tac Q= "nocycle c" in conjI,assumption,drule conjunct2);
 apply (subgoal_tac " (\<exists> E::general_context. a =E\<up>(FutRef fn))\<or>(\<exists> E::general_context. p =E\<up>(FutRef fn))" )
 apply (drule_tac Q= "nocycle c" in conjI,assumption,drule conjunct2);
 apply (simp add: knowsrecI nocycle_def)+;
 apply (erule disjE,(erule exE)+);
 apply (simp add: wf_2a_def);
 apply (rotate_tac 2,erule contrapos_pp)
 apply simp+;
 apply (rule_tac x=A' in bexI)
 apply (rule_tac x=Ea in exI)
 apply (rule_tac x=fn in exI)
 apply (force simp: RequestQueue_def ActiveObject_def);apply force;
 apply (erule exE);
 apply (rotate_tac 3,erule contrapos_pp)
 apply simp+;
 apply (rule_tac x=A' in bexI)
 apply (rule_tac x=fk in bexI)
 apply (rule_tac x="E\<circ>(cCallR (ActRef A') li Ea)" in exI)
 apply (rule_tac x=fn in exI)
 apply (force simp: RequestQueue_def ActiveObject_def);apply force;apply force;

 apply (subgoal_tac " \<exists>E::general_context. Call a li p = E\<up>(FutRef fn)");
 apply (erule exE)
 apply (rotate_tac -1,drule sym)
 apply (frule context_call_1);
 apply (erule disjE)+;
 apply simp+;
 apply force;
 apply (erule disjE)+;
 apply simp+;
 apply (erule bexE);
 apply (case_tac "A=A'");apply simp;
 apply (subgoal_tac "fn\<noteq>fk");apply simp+;

(* REPLY *)
 apply (subgoal_tac "(c(A \<mapsto> (ma(fk \<mapsto> E \<up> ai), a)))=(c(A \<mapsto> ((c.\<^sub>RA)(fk \<mapsto> E \<up> ai), (c.\<^sub>AA))))") (* rewriting for nocycle fut upd *)
 prefer 2; apply (force simp: RequestQueue_def ActiveObject_def);
 apply (rotate_tac -1);
 apply (erule ssubst)

 apply (subgoal_tac "(RefFut fk, RefFut fi) \<in> knows c");
 prefer 2; apply (force simp: knows_def OneRequest_def RequestQueue_def);

 apply (subgoal_tac " \<forall> r. (RefFut fk,r)\<in> knows (c(A \<mapsto> ((c.\<^sub>RA)(fk \<mapsto> E \<up> ai), (c.\<^sub>AA))))\<longrightarrow>(RefFut fk,r)\<in> knows c\<or>(RefFut fi,r)\<in> knows c")

 apply (rule nocycle_fut_upd);
 apply simp+;
 apply force;
 apply (erule contrapos_pp);
 apply (simp add: nocycle_def knowsrecI);
 apply (erule exE,drule minimal_path);
 apply (erule exE,(erule conjE)+);
 apply (erule knowsrec.cases)
 apply simp;
 apply (rotate_tac -3,drule sym,simp+)
 apply (rotate_tac -5)
 apply (drule_tac x="RefFut fk" in spec);
 apply simp+;
 apply (erule disjE);
 apply (rule_tac x="RefFut fk" in exI);
 apply (rule_tac x="[]" in exI);
 apply (force intro: knowsrec.intros)
 apply (rule_tac x="RefFut fk" in exI);
 apply (rule_tac x="[RefFut fi]" in exI);
 apply (force intro: knowsrec.intros)

 (* rec case *)
 apply (erule conjE);  
 apply simp;
 apply (rotate_tac -4,drule sym, simp+)
 apply (drule pre_nocycle_fut_upd');
 apply force;
 apply force;
 apply (drule_tac x=r' in spec);
 apply simp+;
 apply (erule disjE);
 apply (rule_tac x="RefFut fk" in exI);
 apply (rule_tac x="r'#rLb" in exI);
 apply (force intro: knowsrec.intros)
 apply (rule_tac x="RefFut fk" in exI);
 apply (rule_tac x="(RefFut fi)#r'#rLb" in exI);
 apply (force intro: knowsrec.intros)

 (* the first subgoal 1 *)
 apply (rule allI,rule impI); 
 apply (case_tac r);
 apply (subgoal_tac " \<exists> Ea::general_context. E \<up> ai = Ea\<up>(ActRef ActivityRef)")
 apply (erule exE,drule context_inc_Act)
 apply (erule disjE,erule exE,simp);
 apply (rule disjI2);
 apply (force simp: knows_def RequestQueue_def OneRequest_def)
 apply (drule_tac x="FutRef fi" in spec);
 apply (erule exE,simp);
 apply (rule disjI1);
 apply (force simp: knows_def RequestQueue_def OneRequest_def)
 apply (simp add: knows_def RequestQueue_def OneRequest_def)
 apply (erule disjE);apply simp+;
 apply (simp add: disjunct_futures_def RequestQueue_def)
 apply (rotate_tac -1);
 apply (erule bexE);
 apply (subgoal_tac "Aa=A");
 apply simp+;apply (erule conjE,erule exE);
 apply force;

 (* case fut *)
 apply (subgoal_tac " \<exists> Ea::general_context. E \<up> ai = Ea\<up>(FutRef FutureRef)")
 apply (erule exE,drule context_inc_Fut)
 apply (erule disjE,erule exE,simp);
 apply (rule disjI2);
 apply (force simp: knows_def RequestQueue_def OneRequest_def)
 apply (drule_tac x="FutRef fi" in spec);
 apply (erule exE,simp);
 apply (rule disjI1);
 apply (force simp: knows_def RequestQueue_def OneRequest_def)
 apply (simp add: knows_def RequestQueue_def OneRequest_def)
 apply (erule disjE);apply simp+;
 apply (simp add: disjunct_futures_def RequestQueue_def)
 apply (rotate_tac -1);
 apply (erule bexE);
 apply (subgoal_tac "Aa=A");
 apply simp+;apply (erule conjE,erule exE);
 apply force;
done;

lemma nocycle_rtrancl[rule_format]: 
"\<lbrakk> c \<rightarrow>\<^sub>\<zeta>\<^sup>* c' \<rbrakk> \<Longrightarrow> disjunct_futures c \<longrightarrow> wf_conf c \<longrightarrow> nocycle c \<longrightarrow> nocycle c'";
apply (erule_tac P = "% x y. disjunct_futures x \<longrightarrow> wf_conf x \<longrightarrow> nocycle x \<longrightarrow> nocycle y" in rtranclp.induct);
apply simp;
apply (rule impI)+;
apply (erule impE, assumption);
apply (erule impE, assumption);
apply (erule impE, assumption);
apply (rule nocycle_SR);
apply assumption;
apply assumption;
apply (erule disjunct_rtrancl);
apply assumption;
apply (erule wf_conf_rtrancl);
by assumption;


lemma reachable_nocycle: " c\<in> reachable_configs \<Longrightarrow> nocycle c"; 
apply (unfold reachable_configs_def);
apply simp;
apply (erule exE);
apply (erule conjE);
apply (erule nocycle_rtrancl);
apply (rule initial_config_disjunct);
apply (erule wf_init_conf);
apply (erule initial_nocycle);
done;
end